
R_ibs1 <- function(x){
	n <- nrow(x)
	K <- matrix(NA,n,n)
	for(i in 1:n)
		for(j in i:n)
				K[i,j] <- 0.5*mean(2.0 - abs(x[i,]-x[j,]))
				
	for(i in 1:(n-1))
		for(j in (i+1):n)
				K[j,i] <- K[i,j]
	K	
}

R_ibs2 <- function(x1,x2){
	n1 <- nrow(x1)
	n2 <- nrow(x2)
	
	K <- matrix(NA,n1,n2)
	for(i in 1:n1)
		for(j in 1:n2)
				K[i,j] <- 0.5*mean(2.0 - abs(x1[i,]-x2[j,]))
	K	
}

R_ibs <- function(x,x2 = NULL){
	if(!is.null(x2)) return(R_ibs2(x,x2))
	R_ibs1(x)
}

# if 'x' and 'x2' have differing column dimensions, the firts 1..min(ncol(x),ncol(x2)) are used
ibs <- function(x,x2 = NULL){
	storage.mode(x) <- 'double'
	n <- nrow(x)
	if(!is.null(x2)){
		storage.mode(x2) <- 'double'
		n2 <- nrow(x2)	
		return(.C("ibs2_kernel",n,ncol(x),x,n2,ncol(x2),x2,K = matrix(0.0,n,n2),DUP = FALSE)$K)
	}
	.C("ibs_kernel",n,ncol(x),x,K = matrix(0.0,n,n),DUP = FALSE)$K
}

# R's 'euclidean' dist()
edist <- function(x,y = NULL){
	if(is.null(y)) y <- x
	if(!is.matrix(x)) x <- as.matrix(x)
	if(!is.matrix(y)) y <- as.matrix(y)
	storage.mode(x) <- 'double'
	storage.mode(y) <- 'double'
	d <- .C("dist",nrow(x),ncol(x),x,nrow(y),ncol(y),y,dist = double(nrow(x) * nrow(y)),DUP = FALSE)$dist
	dim(d) <- c(nrow(x) , nrow(y))
	d
}

getKernel <- function(x,kernel = c("linear","euclidean","polynomial","rbf","ibs"),para = NULL,X2 = NULL){

	# note: the order of the kernels to be matched against must be exactly as below to match C
	kernel <- substr(tolower(kernel[1]),1,1)
	if(kernel != "l" && is.null(para)) stop("kernel parameter is not set")
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0)-1
	if(kernel == -1) stop("Invalid kernel selected.")

	if(is.null(para)) para <- double(1)
	
	y <- X2
	if(is.null(y)) y <- x
	if(!is.matrix(x)) x <- as.matrix(x)
	if(!is.matrix(y)) y <- as.matrix(y)
	storage.mode(x) <- 'double'
	storage.mode(y) <- 'double'
	K <- .C("getKernel",nrow(x),ncol(x),x,nrow(y),ncol(y),y,
	kernel = as.integer(kernel),para = para,
	K = double(nrow(x) * nrow(y)),DUP = FALSE)$K
	dim(K) <- c(nrow(x) , nrow(y))
	K
}



