##################
## model must contain an intercept
kernlab.W1=function (formula, dat, W, s = 1, lambda = 1, kernel = "linear",
    para = NULL, start.method = "rlogit", eta0.init = NULL,threshold = 1e-4,
    maxit = 100, tol = 1e-06, verbose = F, cost=0, beta.init,arg=0,
	optim.control = control.ipop()
	)
{

	
	verbose <- as.numeric(verbose)
	
    b0.seq=seq(-2,2,0.01)
    time0 = Sys.time()
    tmp = model.frame(formula, dat)
    X = model.matrix(formula, tmp)[, -1, drop = F]  ###
    #y = dat$y
    n = nrow(X)  ###
    TT = dat$TT


    ### get rid of points where W=0
    #W<--(y*TT/sum(TT==1)-y*(1-TT)/sum(TT==0)+cost/n)
    oo<-W!=0
    W<-W[oo]
    X<-as.matrix(X[oo,])
    #y<-y[oo]

	b0.pred <- NA          
	beta.new <- rep(NA,ncol(X) + 1)
	
    Z<-sign(W)
    Wabs<-abs(W)   # absolute value for W
    # Wm<-diag(Wabs)
    #X.diff = get.X.diff(X1, X2)
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" & is.null(para))
        stop("kernel parameter not set")
    #X = rbind(X1, X2)
    K = getK(X, kernel, para)/lambda
    #Q = matrix(0, n1 * n2, n1 * n2)
    # Q = matrix(0, nrow(X), nrow(X))
    # if (!is.double(K))
        # K <- as.double(K)	
    # Zm<-diag(Z)
	# Q = crossprod(Zm, tcrossprod(K,Zm))
	Q <- DXD(Z,K,Z)
	
	# LOQO
	optim.control <- control.ipop(control = optim.control)	
	l. <- rep(0,length(Wabs))
	u. <- Wabs
	r. <- 0    
	A. <- t(as.matrix(Z))
	H. <- Q
	
    if (is.null(eta0.init)) {

        if (kernel == "l"){
          eta.init = Z * (cbind(1,X) %*% beta.init)
        } else
        eta.init=rep(0,sum(oo))
    } else eta.init=eta0.init

   val.init = sum(Wabs*ramp.f(eta.init,s))+lambda*sum(beta.init[-1]^2)/2
   # print(val.approx(eta.init,Z,X,beta.init[1],beta.init[2],Wabs,lambda))
   # beta.seq<-c(beta.init[2],beta.new,beta.new1)
   # val.seq<-val.approx(eta.init,Z,X,beta.init[1],beta.seq,Wabs,lambda)


    iter = 0
    converged = 0
    while (TRUE) {
        iter = iter + 1
        if (iter > maxit) {
            converged = 0
            break
        }
        dh2.deta = ifelse(eta.init < -0.5/s, -s, 0)
        b = s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))
		
		c. <- as.matrix(-0.5*b)
		b. <- -sum(W*dh2.deta)
		
		# LOQO in C in library 'loqo'		
        # fit. <- loqo(c.,H.,A.,b.,l.,u.,r.,
			# solver = optim.control$solver,
			# bound = optim.control$bound,sigf = optim.control$sigf,
			# margin = optim.control$margin,maxiter = optim.control$maxiter,
			# verb = optim.control$verb		
		# )	
		# pp<-loqo::primal(fit.)	
		# converged <- fit.$convergence == 0
		
		# loqo in library 'kernlab'
        fit. <- ipop(c.,H.,A.,b.,l.,u.,r.,
			bound = optim.control$bound,sigf = optim.control$sigf,
			margin = optim.control$margin,maxiter = optim.control$maxiter,
			verb = optim.control$verb		
		)	
		pp<-kernlab::primal(fit.)
		converged <- how(fit.) == 'converged'
		
        if (!converged) print("minQuad does not converge in Iter " %+% iter)
        
		if(any((pp > Wabs) | (pp < 0))){
			opt <- "LOQO"
			warning(sprintf("The solution of the quadratic optimizer %s is out of bounds",opt),immediate. = TRUE)	
		}			
		
        alpha.pred = dh2.deta * Wabs + s * pp

		#cat("alpha_pred ",alpha_pred,"\n")
        if (arg==0){
            b0.pred = 1/Z * (1/2 - Q %*% alpha.pred)
            oo.sp = pp/Wabs>threshold & pp/Wabs<1-threshold

            if (sum(oo.sp)<2) {
                 pp.min<-apply(cbind(pp/Wabs,1-pp/Wabs),1,min)
                 oo.sp=order(pp.min,decreasing=TRUE)[1:5]
				 # if(verbose) cat("threshold free alphas ",(pp/Wabs)[oo.sp],"\n")				 
            }

            b0.pred = mean(b0.pred[oo.sp])
        }

        ##### directly minimizing the loss to get b0
        if (arg==1){
			eta.star=Q %*% alpha.pred

			ETA.seq = tcrossprod(cbind(Z,eta.star) , cbind(b0.seq,1))
				theta=colSums(as.vector(abs(W)) * rp.f(ETA.seq, s, loss=T))
				oo.b0<-(1:length(b0.seq))[theta==min(theta)]
				b0.pred=resample(b0.seq[oo.b0])[1]
        }

        #b0.pred=beta.init[1]
        eta.new = Z * b0.pred + Q %*% alpha.pred
        val.new = sum (Wabs * ramp.f(eta.new,s))
        pen = drop(crossprod(alpha.pred, Q %*% alpha.pred))/2

        if (verbose) cat(iter)
        if (verbose) cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
				
		## Below we do not use beta to check for convergence so 'linear' kernel clause is commented out:
				
        if (kernel == "l") {
            beta.new = c(b0.pred,drop(crossprod(X, alpha.pred * Z)/lambda))
		}
		
		if (verbose)  cat(", loss ", val.new, ", pen ", pen)
		val.new = val.new + pen
		delta = val.new - val.init
		if (abs(delta) < tol) {
			converged = 1
			break
		}
		else {
			val.init = val.new
			eta.init = eta.new
		}
	
        if (verbose)
            cat("\n")
    }
    if (verbose)
        cat("\n")

    Z.pred = Z
    Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
    #linear.combination = Q.pred %*% alpha.pred

    linear.combination = Z.pred * b0.pred + Q.pred %*% alpha.pred
    result = list(converged = converged, iteration = iter, 
		b0=b0.pred,beta = drop(beta.new[-1]),coefficients = drop(beta.new), 
		W=W,X=X,
        alpha.pred = drop(alpha.pred), linear.combination = drop(linear.combination),
        last.minQuad.fit = fit.,
        kernel = kernel, para = para, formula = formula,lambda=lambda)
    # if (kernel == "l") {
        # result$coefficients = drop(t(t(alpha.pred) %*% X.diff)/lambda)
        # result$coefficients = drop(crossprod(X, alpha.pred * Z)/lambda)
        # result$coefficients = beta.new
    # }		
    if (result$converged == 0) {
        result$coefficients = rep(NA, ncol(X) + 1)
        result$alpha.pred = NA
        result$linear.combination = NA
    }
    class(result) = c("rauc", class(result))
    return(result)
}
### this program requires dimension of beta (non-intercept) to be >=2
## model must contain an intercept
kernlab.W2=function (formula, dat, W, s = 1, lambda = 1, kernel = "linear",
    para = NULL, start.method = "rlogit", eta0.init = NULL,
    maxit = 100, tol = 1e-06, threshold = 1e-4, 
	verbose = F, cost=0, beta.init,arg=0,
	optim.control = control.ipop()
	)
{
	verbose <- as.numeric(verbose)
	
	h1.fun<-function(x) {
	 out=0.5-x
	 out=(out>=0)*out
	 return(out)
	}
	h2.fun<-function(x) {
	 out=-0.5-x
	 out=(out>=0)*out
	 return(out)
	}
	h2.der.fun<-function(x){
	 out=-(x<(-0.5))
	 out<-(out!=(-0.5))*out
	 return(out)
	}
	val.approx<-function(eta0,Z,X,b0,beta.seq,Wabs,lambda){
	   eta=tcrossprod(cbind(Z,Z*X) , as.matrix(cbind(b0,beta.seq)))
	   out=apply(Wabs*(h1.fun(eta)-as.vector(h2.fun(eta0))-as.vector(h2.der.fun(eta0))*(eta-as.vector(eta0))),2,sum)+0.5*lambda*beta.seq^2
	 return(out)
	}


	
    b0.seq=seq(-2,2,0.01)
    time0 = Sys.time()
    tmp = model.frame(formula, dat)
    #X1 = model.matrix(formula, tmp[tmp[, 1] == 1, ])[, -1, drop = F]
    #X2 = model.matrix(formula, tmp[tmp[, 1] == 0, ])[, -1, drop = F]
    #n1 = nrow(X1)
    #n2 = nrow(X2)
    #n = n1 + n2

    X = model.matrix(formula, tmp)[, -1, drop = F]  ###
    #y = dat$y
    n = nrow(X)  ###
    TT = dat$TT


    ### get rid of points where W=0
    #W<--(y*TT/sum(TT==1)-y*(1-TT)/sum(TT==0)+cost/n)
    oo<-W!=0
    W<-W[oo]
    X<-as.matrix(X[oo,])
    #y<-y[oo]


	b0.pred <- NA          
	beta.new <- rep(NA,ncol(X) + 1)

    Z<-ifelse(W>0,1,-1)
    Wabs<-abs(W)   # absolute value for W
    #X.diff = get.X.diff(X1, X2)
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" & is.null(para))
        stop("kernel parameter not set")
    #X = rbind(X1, X2)
    K = getK(X, kernel, para)/lambda
	#Q = matrix(0, n1 * n2, n1 * n2)
    # Q = matrix(0, nrow(X), nrow(X))
    # Zm<-diag(Z)
    # Q = Zm %*% K %*% t(Zm)
	Q = DXD(Z,K,Z)
    # Wm<-diag(Wabs)
	# H <- Wm%*%Q%*%Wm
	
	optim.control <- control.ipop(control = optim.control)    
	H. <- DXD(Wabs,Q,Wabs) 
    A. <- t(Z*Wabs)
    l. <- matrix(rep(0,length(Wabs)))
    u. <- matrix(rep(1,length(Wabs)))
    r. <- 0
	
    if (is.null(eta0.init)) {
        if (kernel == "l"){
          eta.init = Z * (cbind(1,X) %*% beta.init)
        } else
        eta.init=rep(0,sum(oo))
    } else eta.init=eta0.init
        # if (!is.null(eta.diff.init)) {
    #        if (start.method == "svm") {
    #            stop("not supported yet")
    #        }
    #        else {
    #            eta.init = eta.diff.init
    #        }
    #    }
    #    else if (!is.null(eta0.init)) {
    #        eta1.init = eta0.init[tmp[, 1] == 1]
    #        eta2.init = eta0.init[tmp[, 1] == 0]
    #        eta.init = eta1.init[rep(1:n1, each = n2)] - eta2.init[rep(1:n2,
    #            n1)]
    #    }
    #    else {
    #        if (is.null(eta0.init)) {
    #            if (kernel == "l") {
    #                if (start.method == "rlogit") {
    #                  fit.rlogit = BYlogreg(formula, dat)
    #                  if (fit.rlogit$convergence) {
    #                    beta.init = coef(fit.rlogit)[-1]
    #                  }
    #                  else {
    #                    beta.init = rep(1, ncol(X1))
    #                  }
    #                }
    #                else if (start.method == "1") {
    #                  beta.init = rep(1, ncol(X1))
    #                }
    #                else if (start.method == "0") {
    #                  beta.init = rep(0, ncol(X1))
    #                }
    #                else stop("start.method not supported: " %+%
    #                  start.method)
    #                if (verbose)
    #                  myprint(beta.init)
    #                eta.init = X.diff %*% beta.init
    #            }
    #            else {
    #                eta.init = rep(0, nrow(X.diff))
    #            }
    #        }
    #    }

    # val.init = mean(Wabs*ramp.f(eta.init, s))
    # val.init = mean(Wabs*ramp.f(eta.init, s))+lambda*beta.init[2]^2/nrow(X)/2
    val.init = sum(Wabs*ramp.f(eta.init,s))+lambda*sum(beta.init[-1]^2)/2
    # print(val.init)
    # print(val.approx(eta.init,Z,X,beta.init[1],beta.init[2],Wabs,lambda))
    # beta.seq<-c(beta.init[2],beta.new,beta.new1)
    # val.seq<-val.approx(eta.init,Z,X,beta.init[1],beta.seq,Wabs,lambda)


    iter = 0
    converged = 0

    while (TRUE) {
        iter = iter + 1
        if (iter > maxit) {
            converged = 0
            break
        }
        dh2.deta = ifelse(eta.init < -0.5/s, -s, 0)
        # b = s^(-2) * (1 - 2 * s * Q %*% dh2.deta)
        # b = s^(-2) * (Wabs - 2 * s * Wm %*% Q %*% (Wabs * dh2.deta))
        # krisz:
		b = s^(-2) * (Wabs - 2 * s * H. %*% dh2.deta)

        c. <- matrix(-0.5*b)
        b. <- sum(-Wabs * Z * dh2.deta)
		
		# LOQO in C in library 'loqo'
        # fit. <- loqo(c.,H.,A.,b.,l.,u.,r.,
			# solver = optim.control$solver,
			# bound = optim.control$bound,sigf = optim.control$sigf,
			# margin = optim.control$margin,maxiter = optim.control$maxiter,
			# verb = optim.control$verb		
		# )	
		# pp<-loqo::primal(fit.)		
		# converged <- fit.$convergence == 0

		# loqo in library kernlab
        fit. <- ipop(c.,H.,A.,b.,l.,u.,r.,
			bound = optim.control$bound,sigf = optim.control$sigf,
			margin = optim.control$margin,maxiter = optim.control$maxiter,
			verb = optim.control$verb		
		)	
		pp<-kernlab::primal(fit.)
		converged <- how(fit.) == 'converged'

        if (!converged)print("minQuad does not converge in Iter " %+% iter)
        
		if(any((pp > Wabs) | (pp < 0))){
			opt <- "LOQO"
			warning(sprintf("The solution of the quadratic optimizer %s is out of bounds",opt),immediate. = TRUE)	
		}			
		
		
        alpha.pred = dh2.deta * Wabs + s * pp * Wabs
       # eta.new = Q %*% alpha.pred

        ##### typical support vector machine way to get b0
        if (arg==0){
            b0.pred = 1/Z * (1/2 - Q %*% alpha.pred)
			oo.sp = pp>threshold & pp<1-threshold
			if (sum(oo.sp)<2) {
				 pp.min<-apply(cbind(pp,1-pp),1,min)
				 oo.sp=order(pp.min,decreasing=TRUE)[1:5]
			}
			b0.pred = mean(b0.pred[oo.sp])
        }
        ##### directly minimizing the loss to get b0
        if (arg==1){
        eta.star=Q %*% alpha.pred

        ETA.seq = tcrossprod(cbind(Z,eta.star) , cbind(b0.seq,1))
            theta=colSums(as.vector(abs(W)) * rp.f(ETA.seq, s, loss=T))
            oo.b0<-(1:length(b0.seq))[theta==min(theta)]
            b0.pred=resample(b0.seq[oo.b0])[1]
        }

        #b0.pred=beta.init[1]
        eta.new = Z * b0.pred + Q %*% alpha.pred
        val.new = sum (Wabs * ramp.f(eta.new,s))
        pen = drop(crossprod(alpha.pred, Q %*% alpha.pred))/2

        if (verbose) cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))

		#### check performance of minQuad ####
		
		## Below we do not use beta to check for convergence so 'linear' kernel clause is commented out:
		
        if (kernel == "l") {
            # beta.new = drop(crossprod(X.diff, alpha.pred)/lambda)
            beta.new = c(b0.pred,drop(crossprod(X, alpha.pred * Z)/lambda))
		}
		val.new = val.new + pen
		delta = val.new - val.init
		if (verbose)  cat(", loss ", val.new, ", pen ", pen," delta ",delta)
		if (abs(delta) < tol) {
			converged = 1
			break
		}
		else {
			val.init = val.new
			eta.init = eta.new
		}
        if (verbose)
            cat("\n")
    }
    if (verbose)
        cat("\n")

    Z.pred = Z
    Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
    #linear.combination = Q.pred %*% alpha.pred

    linear.combination = Z.pred * b0.pred + Q.pred %*% alpha.pred
    result = list(converged = converged, iteration = iter, 
		b0=b0.pred,beta = drop(beta.new[-1]),coefficients = drop(beta.new), 
		W=W,X=X,
        alpha.pred = drop(alpha.pred), linear.combination = drop(linear.combination),
        last.minQuad.fit = fit.,
		# training.auc = fast.auc(drop(linear.combination),
		# c(rep(1, n1), rep(0, n2))),
		# training.theta = caltheta(eta=drop(linear.combination),oo=oo,Y.f=dat$y,TT.f=dat$TT),
		# X = X, y = y, lambda = lambda,
        kernel = kernel, para = para, formula = formula,lambda=lambda)

    # if (kernel == "l") {
        # result$coefficients = drop(t(t(alpha.pred) %*% X.diff)/lambda)
        # result$coefficients = drop(crossprod(X, alpha.pred * Z)/lambda)
        # result$coefficients = beta.new
    # }
	
    if (result$converged == 0) {
        result$coefficients = rep(NA, ncol(X) + 1)
        result$alpha.pred = NA
        result$linear.combination = NA
    }
	class(result) = c("rauc", class(result))
    return(result)
}


