
## From Ying Huang
predict.rtw2<-function (fit, newdata, case.percentage = NULL)
{
    if (inherits(fit,'try-error'))
        return(rep(NA, nrow(newdata)))
    if (is.null(fit$alpha.pred))
        return(rep(NA, nrow(newdata)))
    ff = fit$formula
    tt <- terms(ff)
    ff1 = reformulate(attr(delete.response(tt), "term.labels"))
    X.pred = model.matrix(ff1, newdata, na.action=NULL)[, , drop = F] # krisz: keep intercept
    #Z<-sign(fit$W)
    #Zm<-diag(Z)
    #Z.pred = sign(newdata$W)
    #Zm.pred = diag(Z.pred)
    if (fit$kernel == "l") {
        linear.combination = X.pred %*% coef(fit)
        #linear.combination = fit$b0 + X.pred %*% coef(fit)[-1]
    }
    else {
        K = getK(X.pred, fit$kernel, fit$para, X2 = fit$X)
        linear.combination = fit$b0+crossprod(t(K),fit$alpha.pred*sign(fit$W))/fit$lambda

    }
    res = drop(linear.combination)
    if (!is.null(case.percentage)) {
        threshold = quantile(res, 1 - case.percentage)
        attr(res, "Class") = as.numeric(res > threshold)
    }
    res
}


#### prediction for treatment selection ############
#predict.svml<-function (fit, newdata, case.percentage = NULL)
#{
#    if (is.null(fit$alpha))
#        return(rep(1, nrow(newdata)))
#    #ff = fit$formula
#    #    tt <- terms(ff)
#    #    ff1 = reformulate(attr(delete.response(tt), "term.labels"))
#    #    X.pred = model.matrix(ff1, newdata)[, -1, drop = F]
#    X.pred=newdata$X
#    if (fit$kernel == "l") {
#        #linear.combination = X.pred %*% coef(fit)
#        linear.combination = fit$b0 + X.pred %*% coef(fit)
#    }
#    else {
#        K = getK(X.pred, fit$kernel, fit$para, X2 = fit$SV)
#        linear.combination = fit$b0+crossprod(t(K),fit$y*fit$alpha)
#    }
#    res = drop(linear.combination)
#    if (!is.null(case.percentage)) {
#        threshold = quantile(res, 1 - case.percentage)
#        attr(res, "Class") = as.numeric(res > threshold)
#    }
#    res
#}


predict.svml<-function (fit, newdata, case.percentage = NULL)
{
### this is prediction based on output from svml
    if (inherits(fit,'try-error'))
        return(list(b0=NA,res=rep(NA, nrow(newdata))))
    if (is.null(fit$alpha))
        return(list(b0=NA,res=rep(NA, nrow(newdata))))

        ix <- !(ee(0.0, fit$alpha) | ee(fit$alpha, fit$w * fit$cost))

        K = getK(fit$SV, fit$kernel, fit$para)

        #K = getK(fit$SV, fit$kernel, para=NULL)
        ### note fit$y must be coded as -1,1 here
        bb=1/fit$y-(fit$y*fit$alpha)%*%K
        fit$b0=mean(bb[ix])

         #   datnew=list()
#            datnew$X=cbind(dat.test$x1,dat.test$x2)
#            datnew$W=dat.test$W
#            pred2.0<-predict.svml(fit2.0, newdata=datnew, case.percentage = NULL)
#
#            theta2.0=caltheta(dat.test$yy,dat.test$TT,pred2.0)
#
#            coef2.0=c(fit2.0$b0,coef(fit2.0))

        ff = fit$formula
        tt <- terms(ff)
        ff1 = reformulate(attr(delete.response(tt), "term.labels"))
        X.pred = model.matrix(ff1, newdata)[, -1, drop = F]
        if (fit$kernel == "l") {
            #linear.combination = X.pred %*% coef(fit)
            linear.combination = fit$b0 + X.pred %*% coef(fit)
        } else {
            K = getK(X.pred, fit$kernel, fit$para, X2 = fit$SV)
            linear.combination = fit$b0+crossprod(t(K),fit$y*fit$alpha)
        }
        res = drop(linear.combination)
        if (!is.null(case.percentage)) {
            threshold = quantile(res, 1 - case.percentage)
            attr(res, "Class") = as.numeric(res > threshold)
        }
        list(b0=fit$b0,res=res)
}


predict.svml2<-function (fit, newdata, case.percentage = NULL)
{
### this is prediction based on output from kernlab.SVM2 ?
    if (inherits(fit,'try-error'))
        return(rep(NA, nrow(newdata)))
    if (is.null(fit$alpha))
        return(rep(NA, nrow(newdata)))

        ff = fit$formula
        tt <- terms(ff)
        ff1 = reformulate(attr(delete.response(tt), "term.labels"))
        X.pred = model.matrix(ff1, newdata)[, , drop = F]
        if (fit$kernel == "l") {
            linear.combination = X.pred %*% coef(fit)
            # linear.combination = fit$b0 + X.pred %*% coef(fit)
        } else {
            K = getK(X.pred, fit$kernel, fit$para, X2 = fit$X)
            linear.combination = fit$b0+crossprod(t(K),fit$y*fit$alpha)
        }
        res = drop(linear.combination)
        if (!is.null(case.percentage)) {
            threshold = quantile(res, 1 - case.percentage)
            attr(res, "Class") = as.numeric(res > threshold)
        }
        res
}
