

# lambdas,gammas: 'vec' overwrites the sequence formed as seq(lim[1],lim[2],by = by)
# for linear kernel, set lambdas = list(vec = 1.0,...)
tune.it.large.alter.weights.rts<-function(
formula,wformula = ~w,dat.train, 
formula.tune = yy~TT,wformula.tune = ~weights,dat.tune, 
beta.init,
cost = 0.0,
kernel = c("l","e","p","r","i"),
method = c("rts","svm"),
maxit = 100,tol = 1e-6,threshold = 1e-4,
lambdas = list(vec = NULL,lim = c(.1,.5),by = .01), 
gammas = list(vec = NULL,lim = c(.1,.5),by = .01), 
optim.control = control.loqo(),
verbose = 0
#,arg.tune=0
){

# fit
	y <- model.response(model.frame(formula,dat.train))
	w <- model.matrix(update(wformula,~-1+.),dat.train)
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	if(!INTERCEPT) warning("intercept-free tuning is not implemented for tuning")
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,dat.train)[,-1,drop = FALSE]
	
# prediction	
#	y1 <- model.response(model.frame(formula.pred,dat.tune)); y1 <- .as.double(y1)
	# X1 <- model.matrix(rhs(update(formula,~-1+.)),dat.tune)
	X1 <- model.matrix(formula,dat.tune)[,-1,drop = FALSE]

	if(ncol(X1) != ncol(X)) stop("Column dimension of model matrices for fitting and prediction do not match.")
# caltheta	
	yy  <- model.response(model.frame(formula.tune,dat.tune)); 
	tt1 <- model.matrix(update(formula.tune,~-1+.),dat.tune); 
	w1  <- model.matrix(update(wformula.tune,~-1+.),dat.tune); 

	
	nlambda <- length(lambdas$vec)
	if(nlambda == 0) nlambda <- as.integer(ceiling(diff(lambdas$lim) / lambdas$by) + 1)
	if(!is.finite(nlambda)) stop("error in generating lambda sequence")
	ngamma <- length(gammas$vec)
	if(ngamma == 0) ngamma <- as.integer(ceiling(diff(gammas$lim) / gammas$by) + 1)
	if(!is.finite(ngamma)) stop("error in generating gamma sequence")	
	
    kernel = substr(kernel, 1, 1)	
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0) - 1
	if(kernel < 0) stop("Invalid kernel selected.")
	
	# LOQO control variables are passed to C all as 'double'-s 
	optim.control <- control.loqo(control = optim.control)
	optim.control$solver <- as.integer(match(tolower(optim.control$solver),c("ldl","lu","qr"),nomatch = 1)-1)
	optim.control <- .as.double(unlist(optim.control))

	y <- .as.double(y)
	X <- .as.double(X)
	w <- .as.double(w) 
	X1<- .as.double(X1)
	y1 <- .as.double(yy) #'$yy'
	tt1 <- .as.double(tt1)  #$TT	
	w1 <- .as.double(w1) #'$weights' 
	
	res <- .C("R_tune_it_large_alter_weights",
		ncx = ncol(X),
		nrx = nrow(X),X = X,y = y,w = w,tt = double(1),           # training data 
		nrx1=nrow(X1),X1=X1,y1=y1,w1=w1,tt1=tt1,      # tuning data
		beta.init = as.double(beta.init),
		method = as.integer(match(substr(tolower(method[1]),1,1),c("r","s"),nomatch = 1)-1), #defaults to rts
		threshold = as.double(threshold),
		cost = as.double(cost),
		kernel = as.integer(kernel),
		rts.control = as.double(c(maxit,tol,verbose)),
		optim.control = optim.control,
		ngamma = length(gammas$vec),gamma=as.double(gammas$vec), 
		from.gamma = as.double(gammas$lim[1]),to.gamma = as.double(gammas$lim[2]),by.gamma = as.double(gammas$by),

		nlambda = length(lambdas$vec),lambda=as.double(lambdas$vec), 
		from.lambda = as.double(lambdas$lim[1]),to.lambda = as.double(lambdas$lim[2]),by.lambda = as.double(lambdas$by),
		
		result = double(nlambda * ngamma * 4),
		DUP = FALSE
	)
	result <- res$result
	rm(res)
	dim(result) <- c(nlambda * ngamma,4)
	colnames(result) <- c("lambda","gamma","theta","converged")
	result[,4] <- as.logical(result[,4])
	result
}

R.tune.it.large.alter.weights.rts<-function (formula, dat.train, dat.tune, method, kernel, 
verbose = T,lamcosts=c(.1,.5), gammas=c(.1,.5),
beta.init,varW,arg.tune=0,
maxit = 100,tol = 1e-6,threshold = 1e-4,
optim.control = control.loqo()){

    dat.train$W=dat.train[,varW]
    kernel = substr(kernel, 1, 1)

  dat.train$WW=abs((2*dat.train$TT-1)*dat.train[,varW])
  dat.train$TT.ast=sign(dat.train[,varW])
  dat.train$TT.ast<-ifelse(dat.train$TT.ast==0,1,dat.train$TT.ast)


 if (kernel == "r") {
        params = expand.grid(lamcost = lamcosts, gamma = gammas)
    } else if (kernel == "l") {
        params = expand.grid(lamcost = lamcosts, gamma = gammas[2])
    } else if (kernel == "i") {
        params = expand.grid(lamcost = lamcosts, gamma = gammas[2])
    } else stop("kernel not supported: " %+% kernel)

        res=sapply(1:nrow(params), simplify = "array", function(row.) {
            param = params[row., ]
            if (method == "rts") {
                fit = try.error(kernlab.W1 (formula, dat=dat.train, W=dat.train[,varW], s=1,
					maxit = maxit,tol = tol,threshold = threshold,
					lambda=param$lamcost*nrow(dat.train),
                    kernel=kernel, para=param$gamma, verbose=F,beta.init=beta.init,arg=0,
					optim.control = optim.control))
                pred.rts<-predict.rtw2(fit, newdata=dat.tune, case.percentage = NULL)
                rts.tune=caltheta.weights(dat.tune$yy,dat.tune$TT,pred.rts,dat.tune$weights)
            } else if (method == "svm") {

                #dat.tmp=dat.train[dat.train$WW!=0,]
                #dat.tmp$w.tmp=dat.train[dat.train$WW!=0,]$WW
                #fit.svml=try.error(svml(formula,data=dat.tmp,weights=w.tmp,cost=1/(param$lamcost*sum(abs(dat.tmp$WW))),kernel="linear(l)",class.type='oao'))

                fit.svml = try.error(kernlab.SVW1 (formula, dat=dat.train, W=dat.train[,varW], s=1,
                lambda=param$lamcost*nrow(dat.train),
                    kernel=kernel, para=param$gamma, verbose=F,beta.init=beta.init,arg=0,
					optim.control = optim.control))
                #fit.svml=svml(formula,data=dat[dat$WW!=0,],weights=dat[dat$WW!=0,]$WW,cost=1/(param$lamcost*sum(abs(dat$WW))),kernel="linear(l)",class.type
                #='oao')
                pred.svml<-predict.svml2(fit.svml, newdata=dat.tune, case.percentage = NULL)
                rts.tune=caltheta.weights(dat.tune$yy,dat.tune$TT,pred.svml,dat.tune$weights)
            }
            res = c(lamcost = param$lamcost, gamma = unname(param$gamma),
                tune = rts.tune)
            if (verbose)
                print(res)
            res
        })
    if (arg.tune==0) tmp = res[, which.min(res["tune", ])]
    else if (arg.tune==1) tmp = res[, which.min.random(res["tune", ])]
    else if (arg.tune==2) {
        if (kernel == "l")
        tmp = res[, which.min.mix(res[c("lamcost","tune"), ])]
        if (kernel == "i")
        tmp = res[, which.min.mix(res[c("lamcost","tune"), ])]
        if (kernel == "r")
        tmp = res[, which.min.mix(res[c("lamcost","gamma","tune"), ])]
    }
    if (kernel == "r") {
        res = tmp[c("lamcost", "gamma","tune")]
    } else if (kernel == "l") {
        res = tmp[c("lamcost","tune")]
    } else if (kernel == "i") {
        res = tmp[c("lamcost","tune")]
    } else stop("kernel not supported: " %+% kernel)
    return(res)
}