

rtsw <- function (formula, dat,weights = 1,
	beta.init, eta.init = NULL,s = 1, lambda = 1,
	kernel = "linear",para = NULL,maxit = 100, tol = 1e-06, 
	verbose = FALSE, cost=0,ret.x = FALSE,ret.y=FALSE,
	optim.control = control.loqo()){
	

	w <- weights
	if(length(w) < nrow(dat)) w <- rep(w,length.out = nrow(dat))
	ix <- w != 0
    mf <- model.frame(formula, dat[ix,,drop = FALSE])
	w <- w[ix]
    y <- model.response(mf)

	
	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,mf)[,-1,drop = FALSE]
	
	p <- ncol(X)
	n <- nrow(X)
	m <- 1       # number of constraints

	optim.control <- control.loqo(control = optim.control)
	optim.control$solver <- as.integer(match(tolower(optim.control$solver),c("ldl","lu","qr"),nomatch = 1)-1)
	
	loqo.double.buffer <- .C("R_get_buffer_size_loqo",
	solver = as.integer(optim.control$solver),
	update = as.integer(0),
	m = as.integer(p),n = as.integer(n),
	double.size = integer(1),int.size = integer(1))$double.size
	
	rtw.buffer <- n*n + 9*n + (p+1)
	total.buffer<-loqo.double.buffer + rtw.buffer
		
	if(!length(beta.init)) stop("Please specifiy initial value for beta.")	
	if(length(beta.init) != (ncol(X) + as.numeric(INTERCEPT) )) 
		stop("dimension of beta.init does not match that of model matrix.")	
	if(!INTERCEPT) beta.init <- c(0,beta.init)
	
	.kernel <- kernel
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" && is.null(para))
        stop("kernel parameter not set")
		
	# note the order of characters in 'c("l","e","p","r","i")' must match exactly to the 'enum-type' in C  
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0) - 1
	if(kernel < 0) stop("Invalid kernel selected.")
#	kernel <- max(0,kernel)
	
	if(is.null(para))para <- 0
	
	res <- .C("R_rtsw",
	
# loqo,rts arguments	
	    n = n,p = p,X = X,y = as.double(y),w = as.double(w),
		

# rts arguments		
		kernel = as.integer(kernel),kernel_param = as.double(para),
		s = as.double(s),lambda = as.double(lambda),threshold = as.double(0),
	    intercept = as.integer(INTERCEPT),
		is_eta = as.integer(!is.null(eta.init)) ,eta.init = if(is.null(eta.init)) double(1) else as.double(eta.init),
		is_beta = as.integer(TRUE),beta.init = as.double(beta.init),
# control arguments		
		rts.control = as.double(c(maxit,tol,verbose)),
		
# loqo arguments		
		optim_control = .as.double(unlist(optim.control)),
		buffer_size = as.integer(total.buffer),buffer = double(total.buffer),

# control output		
		convergence = integer(1),

# output
		coefficients = double(p + 1),
		alpha = double(n),
		linear.combination = double(n),
		eta = double(n),
		DUP = FALSE)
		

		
	keep <- c('coefficients','alpha','linear.combination','convergence')
	fit <- res[keep]
	class(fit) <- "rts"
	rm(res)
	fit$converged <- fit$convergence == 0
	
	fit$nSV  <- sum(0.0 < fit$alpha)
	fit$nBSV <- sum(fit$alpha == 1.0)
	fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < 1.0))
	# fit$nSV  <- sum(0.0 < fit$alpha)
	# fit$nBSV <- sum(fit$alpha == w)
	# fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < w))
	
    fit$kernel <- .kernel
	fit$para = para
	fit$formula <- formula
	fit$lambda <- lambda
	fit$x <- X
	fit$y <- y
	fit$w <- w

	names(fit)[names(fit) == 'alpha'] <- 'alpha.pred' # to conform with Ying's pred.rtw2
	
	fit
}

# rtsw with threshold
# implements Ying Huang's kernlab.W1()
rtswt <- function (formula, dat,weights = 1,
	beta.init, eta.init = NULL,s = 1, lambda = 1,threshold = 1e-4,
	kernel = "linear",para = NULL,
	maxit = 100, tol = 1e-06, verbose = FALSE, 
	cost=0,ret.x = FALSE,ret.y=FALSE,
	optim.control = control.loqo()){
	

	w <- weights
	if(length(w) < nrow(dat)) w <- rep(w,length.out = nrow(dat))
	ix <- w != 0
    mf <- model.frame(formula, dat[ix,,drop = FALSE])
	w <- w[ix]
    y <- model.response(mf)
	# if(!setequal(c(0,1),unique(y))) stop("Outcome 'y' should be numeric in {0,1}")
	
	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,mf)[,-1,drop = FALSE]
	

	p <- ncol(X)
	n <- nrow(X)
	m <- 1       # number of constraints

	optim.control <- control.loqo(control = optim.control)
	optim.control$solver <- as.integer(match(tolower(optim.control$solver),c("ldl","lu","qr"),nomatch = 1)-1)
	
	loqo.double.buffer <- .C("R_get_buffer_size_loqo",
	solver = as.integer(optim.control$solver),
	update = as.integer(0),
	m = as.integer(p),n = as.integer(n),
	double.size = integer(1),int.size = integer(1))$double.size
	   
		
	rtw.buffer <- n*n + 9*n + (p+1)
	total.buffer<-loqo.double.buffer + rtw.buffer
		
	if(!length(beta.init)) stop("Please specifiy initial value for beta.")	
	if(length(beta.init) != (ncol(X) + as.numeric(INTERCEPT) )) 
		stop("dimension of beta.init does not match that of model matrix.")	
	if(!INTERCEPT) beta.init <- c(0,beta.init)
	
	.kernel <- kernel
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" && is.null(para))
        stop("kernel parameter not set")	
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0) - 1
	if(kernel < 0) stop("Invalid kernel selected.")
#	kernel <- max(0,kernel)
	if(is.null(para))para <- 0	
	
	res <- .C("R_rtswt",
	
# loqo,rts arguments	
	    n = n,p = p,X = X,y = as.double(y),w = as.double(w),
				
# rts arguments		
		kernel = as.integer(kernel),kernel_param = as.double(para),
		s = as.double(s),lambda = as.double(lambda),threshold = as.double(threshold),
	    intercept = as.integer(INTERCEPT),
		is_eta = as.integer(!is.null(eta.init)) ,eta.init = if(is.null(eta.init)) double(n) else as.double(eta.init),
		is_beta = as.integer(TRUE),beta.init = as.double(beta.init),

# control arguments		
		rts.control = as.double(c(maxit,tol,verbose)),
		
# loqo arguments				
		optim.control = .as.double(unlist(optim.control)),
		buffer_size = as.integer(total.buffer),buffer = double(total.buffer),

# control output		
		convergence = integer(1),

# output
		coefficients = double(p + 1),
		alpha = double(n),
		linear.combination = double(n),
		eta = double(n),
		DUP = FALSE)
		

		
	keep <- c('coefficients','alpha','linear.combination','convergence')
	fit <- res[keep]
	rm(res)
	class(fit) <- "rts"
	fit$converged <- fit$convergence == 0	
	fit$nSV  <- sum(0.0 < fit$alpha)
	fit$nBSV <- sum(fit$alpha == 1.0)
	fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < 1.0))
	# fit$nSV  <- sum(0.0 < fit$alpha)
	# fit$nBSV <- sum(fit$alpha == w)
	# fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < w))	
    fit$kernel <- .kernel
	fit$para = para
	fit$formula <- formula
	fit$lambda <- lambda
	fit$x <- X
	fit$y <- y
	fit$w <- w
	names(fit)[names(fit) == 'alpha'] <- 'alpha.pred' # to conform with Ying's pred.rtw2
	fit
}

# rtsw-svm with threshold
rtswt.svm <- function (formula, dat,weights = 1,
	beta.init, eta.init = NULL,s = 1, lambda = 1,threshold = 1e-4,
	kernel = "linear",para = NULL,
	maxit = 100, tol = 1e-06,verbose = FALSE, 
	cost=0,ret.x = FALSE,ret.y=FALSE,
	optim.control = control.loqo()){
	

	w <- weights
	if(length(w) < nrow(dat)) w <- rep(w,length.out = nrow(dat))
	ix <- w != 0
    mf <- model.frame(formula, dat[ix,,drop = FALSE])
	w <- w[ix]
    y <- model.response(mf)
	
	if(!setequal(c(-1,1),unique(y))){
		warning("Outcome 'y' should be numeric in {-1,1}, trying to recode")
		y <- as.numeric(y)
		y[y == 0] <- -1
		if(!setequal(c(-1,1),unique(y)))
			stop("Could not recode outcome 'y' should to numeric in {-1,1}")
		else
			warning("Successfully recoded outcome 'y' to numeric in {-1,1}")
	}
	
	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,mf)[,-1,drop = FALSE]

	p <- ncol(X)
	n <- nrow(X)
	m <- 1       # number of constraints

	optim.control <- control.loqo(control = optim.control)
	optim.control$solver <- as.integer(match(tolower(optim.control$solver),c("ldl","lu","qr"),nomatch = 1)-1)
	
	loqo.double.buffer <- .C("R_get_buffer_size_loqo",
	solver = as.integer(optim.control$solver),
	update = as.integer(0),
	m = as.integer(p),n = as.integer(n),
	double.size = integer(1),int.size = integer(1))$double.size
	 
		
	rtw.buffer <- 2*n*n + 12*n + (p+1)
	total.buffer<-loqo.double.buffer + rtw.buffer
		
	if(!length(beta.init)) stop("Please specify initial value for beta.")	
	if(length(beta.init) != (ncol(X) + as.numeric(INTERCEPT) )) 
		stop("dimension of beta.init does not match that of model matrix.")	
	if(!INTERCEPT) beta.init <- c(0,beta.init)
	
	.kernel <- kernel
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" && is.null(para))
        stop("kernel parameter not set")	
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0) - 1
	if(kernel < 0) stop("Invalid kernel selected.")
#	kernel <- max(0,kernel)
	if(is.null(para))para <- 0	
	
	res <- .C("R_rtswt_svm",
	
# loqo,rts arguments	
	    n = n,p = p,X = X,y = as.double(y),w = as.double(w),
				
# rts arguments		
		kernel = as.integer(kernel),kernel_param = as.double(para),
		s = as.double(s),lambda = as.double(lambda),threshold = as.double(threshold),
	    intercept = as.integer(INTERCEPT),
		is_eta = as.integer(!is.null(eta.init)) ,eta.init = if(is.null(eta.init)) double(n) else as.double(eta.init),
		is_beta = as.integer(TRUE),beta.init = as.double(beta.init),
# control arguments		
		rts.control = as.double(c(maxit,tol,verbose)),

# loqo arguments				
		optim.control = .as.double(unlist(optim.control)),
		buffer_size = as.integer(total.buffer),buffer = double(total.buffer),

# control output		
		convergence = integer(1),

# output
		coefficients = double(p + 1),
		alpha = double(n),
		linear.combination = double(n),
		eta = double(n),
		DUP = FALSE)
		

		
	keep <- c('coefficients','alpha','linear.combination','convergence')
	fit <- res[keep]
	rm(res) # remove due to .C(DUP = FALSE)
	names(fit)[names(fit) == 'alpha'] <- 'alpha.pred' # to conform with Ying's pred.rtw2
	class(fit) <- "rts"
	fit$converged <- fit$convergence == 0
	
	fit$nSV  <- sum(0.0 < fit$alpha)
	fit$nBSV <- sum(fit$alpha == 1.0)
	fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < 1.0))
	# fit$nSV  <- sum(0.0 < fit$alpha)
	# fit$nBSV <- sum(fit$alpha == w)
	# fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < w))	
    fit$kernel <- .kernel
	fit$para <- para
	fit$formula <- formula
	fit$lambda <- lambda
	fit$x <- X
	fit$y <- y
	fit$w <- w
	fit
}


## algorithm works based on 0 <= alpha[i] <= abs(weights[i])
## hence retain only one copy of nxn matrix Q:
##  Q = Zm %*% K %*% t(Zm)
## optimizer must be able to handle individual bounds on alpha's 
rts.w1<-function (formula, dat,weights = 1,
	beta.init, eta.init = NULL,s = 1, lambda = 1,threshold = 1e-4,  
	kernel = "linear",para = NULL,maxit = 100, tol = 1e-06, 
	verbose = FALSE, cost=0,ret.x = TRUE,ret.y=TRUE,
	use.loqo = TRUE,use.C = TRUE,optim.control = control.loqo())
{

	verbose <- as.numeric(verbose)
	if(use.loqo && !require(loqo))
		stop("Please install library 'loqo'.")
	else if(!require(kernlab))
		stop("Please install library 'kernlab'.")

	kernel = substr(kernel, 1, 1)
    if (kernel != "l" && is.null(para))
        stop("kernel parameter not set")
		
	W <- weights
	if(length(W) < nrow(dat)) W <- rep(W,length.out = nrow(dat))
	ix <-W != 0
    mf <- model.frame(formula, dat[ix,,drop = FALSE])
	W <- W[ix]
    y <- model.response(mf)
    n <- length(y) 
	n1 <- length(y==1)
	n0 <- length(y==0)

	# if(!setequal(c(0,1),unique(y))) stop("Outcome 'y' should be numeric in {0,1}")
	
	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,mf)[,-1,drop = FALSE]
	
	if(!length(beta.init)) stop("Please specifiy initial value for beta.")	
	if(length(beta.init) != (ncol(X) + as.numeric(INTERCEPT) )) 
		stop("dimension of beta.init does not match that of model matrix.")	
	if(!INTERCEPT) beta.init <- c(0,beta.init)

	
    Z<-sign(W)
    Wabs<-abs(W)
    K <- getK(X, kernel, para)/lambda
	Q <- NULL
	if(!use.C){
		Zm<-diag(Z)
		Q <- Zm %*% K %*% t(Zm)
	}else Q <- DXD(Z,K,Z)
    # Q <- Zm %*% K %*% t(Zm) #n x n

	# LOQO
	optim.control <- control.ipop(control = optim.control)	
	l. <- rep(0,length(Wabs))
	u. <- Wabs
	r. <- 0    
	A. <- t(as.matrix(Z))
	H. <- Q
	
    if (is.null(eta.init)) {
        if (kernel == "l"){
          eta.init <- Z * (X %*% beta.init[-1])
		  if(INTERCEPT) eta.init <- eta.init + Z*beta.init[1]
        } else eta.init <- rep(0,n)
    }

    val.init <- sum(Wabs * ramp.f(eta.init,s))
	alpha.pred <- NULL
	b0.pred <- 0
    ############################################
	iter <- 0
    converged<-FALSE
	convergence <- 1
	while(TRUE){
        iter <- iter+1; if(iter > maxit){break;}

        dh2.deta <- ifelse(eta.init < -0.5/s, -s, 0)

        # b <- s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))
        
		#krisz: 
		b <- s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))

		
		b. <- -sum(W * dh2.deta)
		c. <- as.matrix(-0.5*b)
		fit <- NULL
		if(use.loqo){
			if(use.C){
				fit. <- loqo(c.,H.,A.,b.,l.,u.,r.,
					solver = optim.control$solver,
					bound = optim.control$bound,sigf = optim.control$sigf,
					margin = optim.control$margin,maxiter = optim.control$maxiter,
					verb = optim.control$verb		
				)	
			}else{
				fit. <- Rloqo(c.,H.,A.,b.,l.,u.,r.,
					solver = optim.control$solver,
					bound = optim.control$bound,sigf = optim.control$sigf,
					margin = optim.control$margin,maxiter = optim.control$maxiter,
					verb = optim.control$verb		
				)	
			}
			fit$message <- fit.$message
			fit$convergence <- fit.$convergence 
			fit$alpha <- loqo::primal(fit.)
		}else {
			fit. <- ipop(c.,H.,A.,b.,l.,u.,r.,
				bound = optim.control$bound,sigf = optim.control$sigf,
				margin = optim.control$margin,maxiter = optim.control$maxiter,
				verb = optim.control$verb		
			)	
			fit$message <- how(fit.)
			fit$convergence <- as.numeric(!(how(fit.) == 'converged'))
			fit$alpha <- kernlab::primal(fit.)
		}
				
		if(any((fit$alpha > Wabs) | (fit$alpha < 0))){
			opt <- "ipop LOQO";if(use.loqo){opt <- ifelse(use.C,"LOQO in C","LOQO in R")}
			warning(sprintf("The solution of the quadratic optimizer %s is out of bounds",opt),immediate. = TRUE)	
		}
		
        if (fit$convergence != 0)print("minQuad does not converge in Iter " %+% iter)
        if (verbose){ cat(iter," ");cat(fit$message," ")}

        alpha.pred = dh2.deta * Wabs + s * fit$alpha
		
		Qa.pred <- Q %*% alpha.pred
			
        ##### typical support vector machine way to get b0
        if (INTERCEPT){
			# intercept without thresholding
			# fit$nSV  <- sum(0.0 < fit$alpha)
			# fit$nBSV <- sum(fit$alpha == Wabs)
			# fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < Wabs))			
			# if(fit$nFSV > 0){				
				# b0.pred = 1/Z * (1/2 - Qa.pred)
				# b0.pred = mean(b0.pred[(0.0 < fit$alpha) & (fit$alpha < Wabs)])
			# }
			alpha <- fit$alpha / Wabs #alpha on (0,1) scale
			threshold.alpha.free <- (threshold < alpha) & (alpha < 1-threshold) #'free' based on threshold
			if(sum(threshold.alpha.free) < 2){
				threshold.alpha.free <- order(apply(cbind(alpha,1.0-alpha),1,min),decreasing=TRUE)[1:min(5,n)] # take the 5 most extreme,unless n < 5	
			}			
			b0.pred = 1/Z * (1/2 - Qa.pred)
			b0.pred = mean(b0.pred[threshold.alpha.free])
        }
		# if (verbose) cat("b0=",b0.pred,"\n")

        eta.new = Z * b0.pred + Qa.pred
		# cat("\neta_new ",eta.new,"\n")
        val.new = sum(Wabs * ramp.f(eta.new,s))
        pen = 0.5*drop(crossprod(alpha.pred, Qa.pred))

        if (verbose > 1){ 
			cat("beta=",b0.pred)
			if(kernel == "l")cat(drop(crossprod(X, alpha.pred * Z)/lambda))
			cat("\n")
			cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
		}
		
       # check stopping criteria and update eta.init
        # if (kernel=="l"){
            # beta.new = drop(crossprod(X, alpha.pred * Z)/lambda)
            # beta.new = c(b0.pred, beta.new)
            # delta=max(abs((beta.new[-1]-beta.init[-1])/beta.init[-1]))
            # if(verbose) cat (", beta",signif(beta.new,4)," delta ",delta)
# #            if(verbose) cat (", beta[1] ",signif(beta.new[1],4),", beta[-1]/beta[1] ",signif(beta.new[-1]/beta.new[1],4)," delta ",delta)
            # if(delta < tol) {
                # converged = TRUE
                # break;
            # } else {
                # beta.init=beta.new
                # eta.init=eta.new 
            # }
        # } else 
		{
            val.new=val.new + pen
            delta=val.new - val.init
            if(verbose) cat (" delta ",delta)            
            if (abs(delta) < tol) {
                converged <- TRUE
                break;
            } else {
                val.init=val.new
                eta.init=eta.new 
            }
        }		
		if(verbose) cat ("\n")
	}	
	if (verbose) cat ("\n")
	
	# krisz: Why recompute Q as 'Q.pred' ?
    Z.pred <- Z
	Q.pred <- Q
    # Q.pred <- diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
    linear.combination <- Z.pred * b0.pred + Q.pred %*% alpha.pred
	
	
	ret <- NULL
	
	if(kernel == "l"){
		beta.new = drop(crossprod(X, alpha.pred * Z)/lambda)
		beta.new = c(b0.pred, beta.new)
		ret$coefficients <- beta.new
		names(ret$coefficients)[1] <- "(Intercept)"
		ret$beta <- beta.new[-1]
	}
	ret$b0 <- b0.pred
	ret$alpha.pred <- c(alpha.pred)
	ret$linear.combination <- c(linear.combination)
	ret$iterations <- iter
	ret$convergence <- as.numeric(!converged) # as in optim 
	ret$converged <- converged
	ret$delta <- delta
	ret$nSV  <- sum(0.0 < alpha.pred)
	ret$nBSV <- sum(alpha.pred == 1.0)
	ret$nFSV <- sum((0.0 < alpha.pred) & (alpha.pred < 1.0))
	ret$formula <- formula
	if(ret.x) ret$x <- X
	if(ret.y) ret$y <- y
    ret$w <- W
	ret$last.minquad.fit <- fit.	
    ret$kernel <- kernel
	ret$para <- para	
	ret$lambda <- lambda
	ret$s <- s
	
	class(ret) <- "rts"

	if (verbose > 0){ 
		cat("beta=",b0.pred)
		if(kernel == "l")cat(drop(crossprod(X, alpha.pred * Z)/lambda))
		cat("\n")
		cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
	}

	return(ret)	
}

## algorithm works based on 0 <= aw[i] <= 1, aw[i] = alpha[i] / abs(weights[i])
## hence must retain two copies of nxn matrices, Q & H:
##  Q = Zm %*% K %*% t(Zm)
##	H = Wm %*% Q %*% Wm 
rts.w2<-function (formula, dat,weights = 1,
	beta.init, eta.init = NULL,s = 1, lambda = 1, 
	kernel = "linear",para = NULL,maxit = 100, tol = 1e-06, 
	verbose = F, cost=0,
	optim.control = control.loqo())
{

	verbose <- as.numeric(verbose)
	if(!require(kernlab))stop("Please install library 'kernlab'.")

	W <- weights
	if(length(W) < nrow(dat)) W <- rep(W,length.out = nrow(dat))
	ix <-W != 0
    mf <- model.frame(formula, dat[ix,,drop = FALSE])
	W <- W[ix]
    y <- model.response(mf)
    n <- length(y)  
	n1 <- length(y==1)
	n0 <- length(y==0)
	
	# if(!setequal(c(0,1),unique(y))) stop("Outcome 'y' should be numeric in {0,1}")

	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,mf)[,-1,drop = FALSE]

	if(!length(beta.init)) stop("Please specifiy initial value for beta.")	
	if(length(beta.init) != (ncol(X) + as.numeric(INTERCEPT) )) 
		stop("dimension of beta.init does not match that of model matrix.")	
	if(!INTERCEPT) beta.init <- c(0,beta.init)
	
	
    Z<-sign(W)
    Wabs<-abs(W)
	
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" & is.null(para))
        stop("kernel parameter not set")

    K <- getK(X, kernel, para)/lambda
    # Zm<-diag(Z)
    # Q <- Zm %*% K %*% t(Zm) #n x n
	# Wm <- diag(Wabs)
	# H. <- Wm%*%Q%*%Wm
	Q <- DXD(Z,K,Z)	
	
	# LOQO
	optim.control <- control.ipop(control = optim.control)    
	H. <- DXD(Wabs,Q,Wabs)
    A. <- t(Z*Wabs)
    l. <- matrix(rep(0,length(Wabs)))
    u. <- matrix(rep(1,length(Wabs)))
    r. <- 0
	
    if (is.null(eta.init)) {
        if (kernel == "l"){
          eta.init <- Z * (X %*% beta.init[-1])
		  if(INTERCEPT) eta.init <- eta.init + Z*beta.init[1]
        } else eta.init <- rep(0,n)
    }

    val.init <- sum(Wabs * ramp.f(eta.init,s)) + lambda*sum(beta.init[-1]^2)/2

	b0.pred <- 0

    ############################################
	fit <- NULL
	iter <- 0
    converged<-FALSE
	convergence <- 1
	while(TRUE){
		iter <- iter+1; if(iter > maxit){break;}

		dh2.deta <- ifelse(eta.init < -0.5/s, -s, 0)

		# krisz
		b <- s^(-2) * (Wabs - 2 * s * H. %*% dh2.deta)
		# kernlab.W2
        # b = s^(-2) * (Wabs - 2 * s * Wm %*% Q %*% (Wabs * dh2.deta))
		
		c. <- as.matrix(-0.5*b)
		b. <- -sum(W*dh2.deta)
        fit. <- loqo(c.,H.,A.,b.,l.,u.,r.,
			solver = optim.control$solver,
			bound = optim.control$bound,sigf = optim.control$sigf,
			margin = optim.control$margin,maxiter = optim.control$maxiter,
			verb = optim.control$verb		
		)	
		converged <- fit.$convergence == 0
		fit$message <- fit.$message
		fit$convergence <- fit.$convergence 
		fit$alpha <- loqo::primal(fit.)
			
        # fit. <- ipop(c.,H.,A.,b.,l.,u.,r.,
			# bound = optim.control$bound,sigf = optim.control$sigf,
			# margin = optim.control$margin,maxiter = optim.control$maxiter,
			# verb = optim.control$verb		
		# )	
		# fit$alpha<-kernlab::primal(fit.)	
		# converged <- how(fit) == 'converged'

		
		fit$nSV  <- sum(0.0 < fit$alpha)
		fit$nBSV <- sum(fit$alpha == 1.0)
		fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < 1.0))
		
		
		if(any((fit$alpha > Wabs) | (fit$alpha < 0))) warning("Quadratic Optimizer ipop's solution is out of bounds")		
        if (fit$convergence != 0)print("minQuad does not converge in Iter " %+% iter)
        if (verbose){ cat(iter," ")}
		
		alpha.pred = dh2.deta * Wabs + s * fit$alpha * Wabs

		Qa.pred <- Q %*% alpha.pred
		##### typical support vector machine way to get b0
        if (INTERCEPT){
			if(fit$nFSV > 0){
				b0.pred = 1/Z * (1/2 - Qa.pred)
				b0.pred = mean(b0.pred[fit$alpha>0 & fit$alpha<1])
			}
        }

        eta.new = Z * b0.pred + Qa.pred
        val.new = sum(Wabs * ramp.f(eta.new,s))
        pen = 0.5*drop(crossprod(alpha.pred, Qa.pred))

        if (verbose > 1){ 
			cat("beta=",b0.pred)
			if(kernel == "l")cat(drop(crossprod(X, alpha.pred * Z)/lambda))
			cat("\n")
			cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
		}

		
       # check stopping criteria and udpate eta.init
        # if (kernel=="l"){
            # beta.new = drop(crossprod(X, alpha.pred * Z)/lambda)
            # beta.new = c(b0.pred, beta.new)
            # delta=max(abs((beta.new[-1]-beta.init[-1])/beta.init[-1]))
            # if(verbose) cat (", beta",signif(beta.new,4)," delta ",delta)
# #            if(verbose) cat (", beta[1] ",signif(beta.new[1],4),", beta[-1]/beta[1] ",signif(beta.new[-1]/beta.new[1],4)," delta ",delta)
            # if(delta < tol) {
                # converged = TRUE
                # break;
            # } else {
                # beta.init=beta.new
                # eta.init=eta.new 
            # }
        # } else 
		{
            val.new=val.new + pen
            delta=val.new - val.init
			if (verbose)  cat(", loss ", val.new, ", pen ", pen," delta ",delta)
            if (abs(delta) < tol) {
                converged <- TRUE
                break;
            } else {
                val.init=val.new
                eta.init=eta.new 
            }
        }	
		if (verbose) cat ("\n")		
	}	
	
	if (verbose) cat ("\n")
	if(kernel == "l"){
		beta.new = drop(crossprod(X, alpha.pred * Z)/lambda)
		beta.new = c(b0.pred, beta.new)	
		fit$coefficients <- beta.new
		names(fit$coefficients)[1] <- "(Intercept)"
		fit$beta <- beta.new[-1]
	}

	fit$b0 <- b0.pred
	fit$alpha.pred <- c(alpha.pred)
	fit$iterations <- iter
	fit$convergence <- as.numeric(!converged) # as in optim 
	fit$converged <- converged
	fit$nSV  <- sum(0.0 < alpha.pred)
	fit$nBSV <- sum(alpha.pred == 1.0)
	fit$nFSV <- sum((0.0 < alpha.pred) & (alpha.pred < 1.0))
	fit$formula <- formula
	fit$x <- X
	fit$y <- y
    fit$w <- w	
	fit$last.minquad.fit <- fit.
    fit$kernel <- kernel
	fit$para = para	
	class(fit) <- "rts"	
	
	if (verbose > 0){ 
		cat("beta=",b0.pred)
		if(kernel == "l")cat(drop(crossprod(X, alpha.pred * Z)/lambda))
		cat("\n")
		cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
	}
	
	return(fit)	
}
