#include "matrix.h"
#include <R.h>
#include <string.h>

// dgetru1: dgetr rank-1 update of the dgetrf() LU-factorization of 'A' <- A + uv'
// _diag  : the implementation below only does a rank-1 update s.t. uv' has exactly one nonzero entry on its diagonal !

// NOTE : The Fortran 'dlu1up driver (vs dlup1up) does not accept the row-permutation vector corresponding 
// to 'ipiv' returned by LAPACK, therefore we are stuck with the initial pivoting of the augmented KKT matrix 
// even as the diagonals change ! Thus after repeated calls to this update function, we may be numerically unstable.

void C_dgetru1_diag(int m,int n,double* dgetrf_LU,double* L,double* U,double* diagonal,int* ipiv){ 
	int i,j;
	int m_plus_n = m + n;
	int _perm[m_plus_n];int* perm = &_perm[0];
	int _iperm[m_plus_n];;int* iperm = &_iperm[0];	
	double _diag[m_plus_n];double* diag = &_diag[0];
	double _u[m_plus_n];double* u = &_u[0];
	double _v[m_plus_n];double* v = &_v[0];

	
	for(i = 0;i < m_plus_n;i++)diag[i] = 1.0; // set diagonal of L to 1.0
	lower_trap(m_plus_n,m_plus_n,dgetrf_LU,diag,0,L);
	upper_trap(m_plus_n,m_plus_n,dgetrf_LU,NULL,-1,U);

	rowperm_ipiv(&m_plus_n,ipiv,perm);
	invperm(&m_plus_n,perm,iperm);

	for(int k = 0;k < m_plus_n;k++){

		// one of {u,v} is destroyed in dlu1up
		memset(u,0,(size_t)(m_plus_n * sizeof(double)));
		memset(v,0,(size_t)(m_plus_n * sizeof(double)));

		// P %*% u is just a switch w.r.t. the 'perm' if 'u' has 0's everywhere except one place 
		// but 'iperm' is needed since A = P * L * U in LAPACK is P*A = L*U
		u[iperm[k]-1] = diagonal[k]; 
		v[k] = 1.0;

		F77_CALL(dlu1up)(&m_plus_n,&m_plus_n,L,&m_plus_n,U,&m_plus_n,u,v);
	
		// // P u v' = (Pu)v' = _u v'
		// // if uv' is a diagonal matrix, otherwise u <- row.PL(perm) %*% u
		// if(diagonal){
			// for(i = 0;i < m;i++)
				// if(u[i] != 0.0) 
					// _u[iperm[i]-1] = u[i];
		// }else{
			// u <- row.PL(perm) %*% u
		// }				
		// F77_CALL(dlu1up)(&m,&n,L,&m_plus_n,U,&m_plus_n,&_u[0],&v[0]);	
	}
		

	// no need to set diagonal of 'L' to zero because we have i > j		
	// for(i = 0;i < m_plus_n;i++) L[i * m_plus_n + i] = 0.0; // set diag(L) <- 0.0
	for(i = 0;i < m_plus_n;i++){
		for(j = 0;j < m_plus_n;j++){
			int kx = i + j*m_plus_n;
			if(i > j) dgetrf_LU[kx] = L[kx]; else  dgetrf_LU[kx] = U[kx];
			//dgetrf_LU[kx] = L[kx] + U[kx];	
		}
	}
	
}
