#include "rts.h"
#include "matrix.h"
#include "kernels.h"
#include <string.h>
#include <stdlib.h>

#define	MAX(A,B)	((A) > (B) ? (A) : (B))
#define	MIN(A,B)	((A) < (B) ? (A) : (B))
#define	ABS(x)  	((x) > 0 ? (x) : (-(x)))
#define	SIGN(x)	    ((x) > 0 ? 1 : (x)==0 ? 0 : -1) 
#define EQ(x,y,eps) ((ABS((x)-(y))) <= (eps) ? 1 : 0) 

// internal_buffer_size > (*nrx) * (*nrx1) + (*nrx1) otherwise allocated
void predict_rtsw(RTS_FIT* fit,int nrx,double* x,double* result)
{
	int i;
	int ione = 1;
	int n1 = nrx;
	int n  = fit->data.nrx;
	int p = fit->data.ncx;
	double* linear_combination = result; // prediction output of length nrx1
	double lambda = fit->param.lambda;
	RTS_PARAM* param = &fit->param;
	KERNEL_ARG karg;memset(&karg,0,sizeof(KERNEL_ARG));
	karg.x1 = fit->data.X;karg.x2 = fit->data.X;karg.nr1 = n;karg.nr2 = n;karg.nc1 = p;karg.nc2 = p;

	if (fit->kernel.type == LINEAR){
        // linear.combination = fit$b0 + X.pred %*% coef(fit)
		// matprod(x,&nrx,&fit->ncx,fit->beta + 1,&fit->ncx,&ione,linear_combination);
		matprod(x,&n1,&p,fit->beta + 1,&p,&ione,linear_combination);
		if(fit->beta[0] != 0.0){
			for(i = 0;i < n1;i++) 
				linear_combination[i] += fit->beta[0];		
		}
		return;
	}


	double* buffer = NULL;
	int internal_buffer_size = n1*n + n;
	int internal_buffer = 1;
	if(param->aux_buffer && param->aux_buffer_size)
		internal_buffer = internal_buffer_size > param->aux_buffer_size;
	if(internal_buffer){
		buffer = (double*)calloc((size_t)internal_buffer_size , sizeof(double));  
	}else{
		buffer = param->aux_buffer;
		memset(buffer,0,(size_t)internal_buffer_size * sizeof(double));
	}
    double* _p = buffer;
	double* K = _p;_p+= n1 * n;
	double* Z_alpha = _p;

	for(i = 0;i < n;i++) Z_alpha[i] = SIGN(fit->data.w[i]) * fit->alpha_pred[i];
	//int kernel = (int)fit->kernel;getKernel(&n1,&p,x,&n,&p,fit->X,&kernel,&fit->kernel_param,K);	// K_{nrx1 x nrx}
	// matprod(K,&n1,&n,y_alpha,&f,&ione,linear_combination);
	getK(&karg,&fit->kernel,K);
	matprod(K,&n1,&n,Z_alpha,&n,&ione,linear_combination);
    for(i = 0;i < nrx;i++)linear_combination[i] /= lambda;
	if(fit->beta[0] != 0.0){
		for(i = 0;i < n1;i++) 
			linear_combination[i] += fit->beta[0];		
	}

	if(internal_buffer)free(buffer);
}

void R_predict_rtsw2(
int* ncx,
int* nrx1,double* X1,                 // newdata_(nrx1 x ncx) on which prediction is based
int* nrx,double* X,                   // data_(nrx x ncx) used in fit 
double* y,
double* w,                            // (+/-)weights_{nrx x 1} used in fitting
double* alpha,						  // estimated alpha_{nrx x 1} from fitting
double* beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
int* kernel,double* kernel_param,     // kernel and its parameter 
double* lambda,                      // tuning parameters
int* aux_buffer_size,double* aux_buffer,     // buffer needed for calculation
double* linear_combination            // prediction output of length nrx1
){
	RTS_FIT fit;
	memset(&fit,0,sizeof(RTS_FIT));
	fit.data.ncx = *ncx;
	fit.data.nrx = *nrx;
	fit.data.X = X;
	fit.data.y = y;
	fit.data.w = w;
	fit.alpha_pred = alpha;
	fit.beta = beta;
	fit.param.aux_buffer_size = *aux_buffer_size;
	if(fit.param.aux_buffer_size) fit.param.aux_buffer = aux_buffer;
	fit.kernel.type = (KERNEL_TYPE)*kernel;
	fit.kernel.param = *kernel_param;		
	fit.param.lambda = *lambda;
	fit.param.s = 1;	
	predict_rtsw(&fit,*nrx1,X1,linear_combination);
}

// internal_buffer_size > (*nrx) * (*nrx1) + (*nrx1) otherwise allocated
void predict_svmw(RTS_FIT* fit,int nrx,double* x,double* result)
{
	int i;
	int ione = 1;
	int n1 = nrx;
	int n  = fit->data.nrx;
	int p = fit->data.ncx;
	double* linear_combination = result; // prediction output of length nrx1
	double lambda = fit->param.lambda;
	RTS_PARAM* param = &fit->param;
	KERNEL_ARG karg;memset(&karg,0,sizeof(KERNEL_ARG));
	karg.x1 = fit->data.X;karg.x2 = fit->data.X;karg.nr1 = n;karg.nr2 = n;karg.nc1 = p;karg.nc2 = p;
	if (fit->kernel.type == LINEAR){
        // linear.combination = fit$b0 + X.pred %*% coef(fit)
		matprod(x,&n1,&p,fit->beta + 1,&p,&ione,linear_combination);
		if(fit->beta[0] != 0.0){
			for(i = 0;i < n1;i++) 
				linear_combination[i] += fit->beta[0];		
		}
		return;
	}

	
	double* buffer = NULL;
	int internal_buffer_size = n1*n + n;
	int internal_buffer = 1;
	if(param->aux_buffer && param->aux_buffer_size)
		internal_buffer = internal_buffer_size > param->aux_buffer_size;
	if(internal_buffer){
		buffer = (double*)calloc((size_t)internal_buffer_size , sizeof(double));  
	}else{
		buffer = param->aux_buffer;
		memset(buffer,0,(size_t)internal_buffer_size * sizeof(double));
	}
    double* _p = buffer;
	double* K = _p;_p+= n1 * n;
	double* y_alpha = _p;

	for(i = 0;i < n;i++) y_alpha[i] = fit->data.y[i] * fit->alpha_pred[i];
	//int kernel = (int)fit->kernel;getKernel(&n1,&p,x,&n,&p,fit->X,&kernel,&fit->kernel_param,K);	// K_{nrx1 x nrx}
	// matprod(K,&n1,&n,y_alpha,&f,&ione,linear_combination);
	getK(&karg,&fit->kernel,K);
	matprod(K,&n1,&n,y_alpha,&n,&ione,linear_combination);
	if(fit->beta[0] != 0.0){
		for(i = 0;i < n1;i++) 
			linear_combination[i] += fit->beta[0];		
	}

	if(internal_buffer)free(buffer);
}

void R_predict_svml2(
int* ncx,
int* nrx1,double* X1,                 // newdata_(nrx1 x ncx) on which prediction is based
int* nrx,double* X,                   // data_(nrx x ncx) used in fit 
double* y,
double* w,                            // (+/-)weights_{nrx x 1} used in fitting
double* alpha,						  // estimated alpha_{nrx x 1} from fitting
double* beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
int* kernel,double* kernel_param,     // kernel and its parameter 
double* lambda,                      // tuning parameters
int* aux_buffer_size,double* aux_buffer,     // buffer needed for calculation
double* linear_combination            // prediction output of length nrx1
){
	RTS_FIT fit;
	memset(&fit,0,sizeof(RTS_FIT));
	fit.data.ncx = *ncx;
	fit.data.nrx = *nrx;
	fit.data.X = X;
	fit.data.y = y;
	fit.data.w = w;
	fit.alpha_pred = alpha;
	fit.beta = beta;
	fit.param.aux_buffer_size = *aux_buffer_size;
	if(fit.param.aux_buffer_size) fit.param.aux_buffer = aux_buffer;
	fit.kernel.type = (KERNEL_TYPE)*kernel;
	fit.kernel.param = *kernel_param;
	fit.param.lambda = *lambda;
	fit.param.s = 1;
	predict_rtsw(&fit,*nrx1,X1,linear_combination);
}

// predict.rtw2<-function (fit, newdata, case.percentage = NULL)
// {
    // if (inherits(fit,'try-error'))
        // return(rep(NA, nrow(newdata)))
    // if (is.null(fit$alpha.pred))
        // return(rep(NA, nrow(newdata)))
    // ff = fit$formula
    // tt <- terms(ff)
    // ff1 = reformulate(attr(delete.response(tt), "term.labels"))
    // X.pred = model.matrix(ff1, newdata, na.action=NULL)[, -1, drop = F]
    // #Z<-sign(fit$W)
    // #Zm<-diag(Z)
    // #Z.pred = sign(newdata$W)
    // #Zm.pred = diag(Z.pred)
    // if (fit$kernel == "l") {
        // #linear.combination = X.pred %*% coef(fit)
        // linear.combination = fit$b0 + X.pred %*% coef(fit)
    // }
    // else {
        // K = getK(X.pred, fit$kernel, fit$para, X2 = fit$X)
        // linear.combination = fit$b0+crossprod(t(K),fit$alpha.pred*sign(fit$W))/fit$lambda

    // }
    // res = drop(linear.combination)
    // if (!is.null(case.percentage)) {
        // threshold = quantile(res, 1 - case.percentage)
        // attr(res, "Class") = as.numeric(res > threshold)
    // }
    // res
// }



// predict.svml2<-function (fit, newdata, case.percentage = NULL)
// {
// ### this is prediction based on output from kernlab.SVM2
    // if (inherits(fit,'try-error'))
        // return(rep(NA, nrow(newdata)))
    // if (is.null(fit$alpha))
        // return(rep(NA, nrow(newdata)))

        // ff = fit$formula
        // tt <- terms(ff)
        // ff1 = reformulate(attr(delete.response(tt), "term.labels"))
        // X.pred = model.matrix(ff1, newdata)[, -1, drop = F]
        // if (fit$kernel == "l") {
            // #linear.combination = X.pred %*% coef(fit)
            // linear.combination = fit$b0 + X.pred %*% coef(fit)
        // } else {
            // K = getK(X.pred, fit$kernel, fit$para, X2 = fit$X)
            // linear.combination = fit$b0+crossprod(t(K),fit$y*fit$alpha)
        // }
        // res = drop(linear.combination)
        // if (!is.null(case.percentage)) {
            // threshold = quantile(res, 1 - case.percentage)
            // attr(res, "Class") = as.numeric(res > threshold)
        // }
        // res
// }
