#include "rts.h"
#include "matrix.h"
#include <R.h>
#include <stdlib.h>

// initialize to default control setting, can add further arguments
void init_control_rts(RTS_CONTROL* control){
	if(!control)return;
	memset(control,0,sizeof(RTS_CONTROL));
	control->maxit = 100;
	control->tol = 1e-6;
	control->verbose = 0;
}
void print_control_rts(RTS_CONTROL* control){
	Rprintf("rts control: ");
	if(!control){
		Rprintf("\n");
		return;
	}
	Rprintf("maxit(%i) tol(%.3e) verbose(%i)\n",control->maxit,control->tol,control->verbose);
}
// initialize to default param setting, can add further arguments
void init_param_rts(RTS_PARAM* param){
	if(!param)return;
	memset(param,0,sizeof(RTS_PARAM));
	param->is_intercept = 1;
	param->lambda = 1;
	param->s = 1;
	param->threshold = 0;	
	init_control_loqo(&param->loqo_control);
	init_control_rts(&param->control);
}

void print_param_rts(RTS_PARAM* param){
	Rprintf("rts param:\n");
	if(!param)return;
	Rprintf("estimate intercept yes(1)/no(0):%i\n",param->is_intercept);
	Rprintf("lambda(%.3e) s(%.3e) threshold(%.3e)\n",param->lambda,param->s,param->threshold);
	print_control_loqo(&param->loqo_control);
	print_control_rts(&param->control);
	if(param->beta_init){
		Rprintf("beta_init =");for(int i = 0;i < param->nbeta;i++)Rprintf("%+.3e ",param->beta_init[i]);Rprintf("\n");
	}
	if(param->eta_init){
		Rprintf("eta_init  =");for(int i = 0;i < param->neta;i++)Rprintf("%+.3e ",param->eta_init[i]);Rprintf("\n");
	}	Rprintf("\n");
}

void print_mf_rts(RTS_MODEL_FRAME* mf,int print_X){
	Rprintf("rts model frame:\n");
	if(!mf)return;
	if(mf->y){Rprintf("y =");for(int i = 0;i < mf->nrx;i++)Rprintf("%+.3e ",mf->y[i]);Rprintf("\n");}
	if(mf->w){Rprintf("w =");for(int i = 0;i < mf->nrx;i++)Rprintf("%+.3e ",mf->w[i]);Rprintf("\n");}
	if(mf->tt){Rprintf("tt=");for(int i = 0;i < mf->nrx;i++)Rprintf("%+.3e ",mf->tt[i]);Rprintf("\n");}
	if(mf->X){if(print_X)print_matrix(mf->nrx,mf->ncx,mf->X);}
	Rprintf("\n");
}


 // caltheta.weights<-function(Y,TT,score,weights){
 // if (sum(!is.na(score))==0) out=NA
 // else {
 // out=sum(Y*(score>0)*TT*weights,na.rm=T)/sum(TT*weights,na.rm=T)+sum(Y*(score<=0)*(1-TT)*weights,na.rm=T)/sum((1-TT)*weights,na.rm=T)
 // }
 // return(out)
// }
// caltheta<-function(Y,TT,score){
 // out=mean((Y*(score>0))[TT==1])+mean((Y*(score<=0))[TT==0])
 // return(out)
// }

void rts_caltheta_weights(int* n,double* Y,double* weights,double* TT,double* score,double* cost,double* theta){
  
	double num = 0.0;
	double denum = 0.0;
	for(int i = 0;i < *n;i++){
		double TTw = TT[i] * weights[i];
		num += Y[i] * (double)(score[i] > 0.0) * TTw;
		denum += TTw;
	}
	*theta = num / denum;

	num = 0.0;
	denum = 0.0;
	for(int i = 0;i < *n;i++){
		double TTw = (1.0-TT[i]) * weights[i];
		num += Y[i] * (double)(score[i] <= 0.0) * TTw;
		denum += TTw;
	}
	*theta += num / denum;

	if(*cost != 0.0){
		num = 0.0;
		denum = 0.0;
		for(int i = 0;i < *n;i++){
			num += (double)(score[i] > 0.0)  * weights[i];
			denum += weights[i];
		}
		//Rprintf("out=%f cost=%f ssw=%f sw=%f \n",*theta,*cost,num,denum);
		*theta += (*cost) * num / denum;
	}
}
