#include "rts.h"
#include <stdlib.h>
#include <string.h>
#include <R.h>

#define	MAX(A,B)	((A) > (B) ? (A) : (B))
#define	MIN(A,B)	((A) < (B) ? (A) : (B))
#define	ABS(x)  	((x) > 0 ? (x) : (-(x)))
#define	SIGN(x)	    ((x) > 0 ? 1 : (x)==0 ? 0 : -1) 
#define EQ(x,y,eps) ((ABS((x)-(y))) <= (eps) ? 1 : 0) 
#define CX(i,j,ld)  (((j)*(ld))+(i)) //column-major index
#define RX(i,j,ld)  (((i)*(ld))+(j)) //row-major index



void (*fit)(
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta);

void (*predict)(
int* ncx,
int* nrx1,double* X1,                 // newdata_(nrx1 x ncx) on which prediction is based
int* nrx,double* X,                   // data_(nrx x ncx) used in fit 
double* y,
double* w,                            // (+/-)weights_{nrx x 1} used in fitting
double* alpha,						  // estimated alpha_{nrx x 1} from fitting
double* beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
int* kernel,double* kernel_param,     // kernel and its parameter 
double* lambda,                      // tuning parameters
int* aux_buffer_size,double* aux_buffer,     // buffer needed for calculation
double* linear_combination            // prediction output of length nrx1
);




void R_tune_it_large_alter_weights(
int* ncx,
int* nrx,double* X,double* y,double* w,double* tt,           // training data 
int* nrx1,double* X1,double* y1,double* w1,double* tt1,      // tuning data
double* beta_init,
int* _method,
double* threshold,
double* cost,
int* kernel,
double* rts_control,// maxit,tol,verbose,
double* optim_control,
int* _n_kernel_param,double* _kernel_param, 
double* from_kernel_param,double* to_kernel_param,double* by_kernel_param,
int* _n_lambda,double* _lambda, 
double* from_lambda,double* to_lambda,double* by_lambda,
double* result //_{nlambda * ngamma x 3}
){

	int i,j;
	int n_kernel_param = *_n_kernel_param;
    int n_lambda = *_n_lambda;
	double lambda = 1.0;
	double gamma = 0.0; //kernel_param

	// int maxit = (int)rts_control[0];
	// double tol = rts_control[1];
	int verbose = (int)rts_control[2];
	
	// RTS_FIT fit;
	// memset(&fit,0,sizeof(RTS_FIT));
	// RTS_FIT* arg = &fit;
	
	int n = *nrx;
	int n1 = *nrx1;
	int buffer_size = n1*n + 3*n + (*ncx + 1);
	double* buffer = (double*)calloc((size_t)buffer_size , sizeof(double)); 
	if(!buffer){
		Rprintf("Error allocating buffer");
		return;
	}
	double* _p = buffer;
	double* beta = _p;_p+=(*ncx + 1);
	double* eta = _p;_p+=n;
	double* alpha = _p;_p+=n;
	double* linear_combination = _p;_p+=n;
	double* lin_comb_pred = _p;_p+=n1;
	
	int method = 0;if(_method)method = *_method;
	if(method == 0){
		fit = R_rtswt;
		predict = R_predict_rtsw2;
		if(verbose > 0)Rprintf("Tuning method 'rts' %i\n",method);
	}else{
		fit = R_rtswt_svm;
		predict = R_predict_svml2;	
		if(verbose > 0)Rprintf("Tuning method 'svm' %i\n",method);
	}
		
	
	// a = 1,b = 3,s = .3
	// a + (0:(n-1))*s = 1.0 1.3 1.6 1.9 2.2 2.5 2.8
    // a + (0 : n ) *s = 1.0 1.3 1.6 1.9 2.2 2.5 2.8 3.1 so take floor() of last element, done by MIN() here
	if(n_lambda == 0)
		n_lambda = 1 + (int)ceil((*to_lambda - *from_lambda) / *by_lambda);
	
	if(n_kernel_param == 0)
		n_kernel_param = 1 + (int)ceil((*to_kernel_param - *from_kernel_param) / *by_kernel_param);
		
	int NROW = n_lambda * n_kernel_param;
	int NCOL = 4;
	memset(result,0,NROW * NCOL * sizeof(double));
	int kx = 0;//indexes rows of result[]
	
	for(j = 0;j < n_kernel_param;j++){
	
		if(*_n_kernel_param){
			gamma = _kernel_param[j];
		}else{
			gamma = from_kernel_param[0] + (double)(j) * by_kernel_param[0];
			gamma = MIN(gamma,to_kernel_param[0]);// could happen only when j = _n - 1			
		}
			
		for(i = 0;i < n_lambda;i++){
			
			if(*_n_lambda){
				lambda = _lambda[i];
			}else{
				lambda = from_lambda[0] + (double)(i) * by_lambda[0];
				lambda = MIN(lambda,to_lambda[0]);// could happen only when i = _n - 1
			}
			lambda *= (double)n;
			
/// fit				
			int is_intercept = 1;
			int is_eta = 0;
			int is_beta = 1;
			int status = 1; // 0 indicates 'ok'
			// int verb = verbose - 1;
			double s = 1;
			fit(
				nrx,ncx,X,y,w,
				kernel,&gamma,
				&s,&lambda,
				threshold,
				&is_intercept,
				&is_eta,NULL,//eta_init,
				&is_beta,beta_init,
				rts_control,   // maxit,tol,&verb,
				optim_control,
				NULL,NULL,     //int* _external_buffer_size,double* _buffer,
				&status,
				beta,
				alpha,
				linear_combination,
				eta
			);
			if(status){
				if(verbose > 0){
					Rprintf("%i fitting diverged for (lambda,gamma) = (%.8f,%.8f) in quadratic optimizer LOQO \n",
					kx+1,lambda / (double)n,gamma);
				}
				
				// // if did not converge and we just want to move on
				// kx++;
				// continue; 
			}
	
			
/// predict
			int zero = 0;
			predict(
				ncx,
				nrx1,X1,                 // newdata_(nrx1 x ncx) on which prediction is based
				nrx,X,                   // data_(nrx x ncx) used in fit 
				y,
				w,                            // (+/-)weights_{nrx x 1} used in fitting
				alpha,						  // estimated alpha_{nrx x 1} from fitting
				beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
				kernel,&gamma,     // kernel and its parameter 
				&lambda,                      // tuning parameters
				&zero,NULL,     // buffer needed for calculation
				lin_comb_pred            // prediction output of length nrx1
			);		
			double theta;
			rts_caltheta_weights(nrx1,y1,w1,tt1,lin_comb_pred,cost,&theta);

			if((verbose == 1) && ((KERNEL_TYPE)*kernel == LINEAR)){
				Rprintf("beta ");
				for(int k = 0;k < (1 + *ncx);k++)Rprintf("%.6f ",beta[k]);
				Rprintf("\n");
			}
			if(verbose > 0)Rprintf("%i (lambda,gamma,theta) = (%.8f,%.8f,%.8f)\n\n",kx+1,lambda / (double)n,gamma,theta);

			result[CX(kx,0,NROW)] = lambda / (double)n;
			result[CX(kx,1,NROW)] = gamma;
			result[CX(kx,2,NROW)] = theta;
			result[CX(kx,3,NROW)] = (double)(status == 0 ? 1 : 0);
			kx++;
		}
	}
	free(buffer);
}	

// R function implemented above
// tol = 1e-6 below
// tune.it.large.alter.weights<-function (formula, dat.train, dat.tune, method = c("rts,svm"), 
// kernel, verbose = T,lamcosts=c(.1,.5), gammas=c(.1,.5), beta.init,varW,arg.tune=0){

    // dat.train$W=dat.train[,varW]
    // kernel = substr(kernel, 1, 1)

  // dat.train$WW=abs((2*dat.train$TT-1)*dat.train[,varW])
  // dat.train$TT.ast=sign(dat.train[,varW])
  // dat.train$TT.ast<-ifelse(dat.train$TT.ast==0,1,dat.train$TT.ast)


 // if (kernel == "r") {
        // params = expand.grid(lamcost = lamcosts, gamma = gammas)
    // } else if (kernel == "l") {
        // params = expand.grid(lamcost = lamcosts, gamma = gammas[2])
    // } else stop("kernel not supported: " %+% kernel)

        // res=sapply(1:nrow(params), simplify = "array", function(row.) {
            // param = params[row., ]
            // if (method[1] == "rts") {
                // fit = try.error(kernlab.W1 (formula, dat=dat.train, W=dat.train[,varW], s=1, lambda=param$lamcost*nrow(dat.train),
                    // kernel=kernel, para=param$gamma, verbose=F,beta.init=beta.init,arg=0,tol = 1e-6))
                // pred.rts<-predict.rtw2(fit, newdata=dat.tune, case.percentage = NULL)
                // rts.tune=caltheta(dat.tune$yy,dat.tune$TT,pred.rts)
				// if (verbose){
					// if(kernel == "l")cat("beta=",c(fit$b0,fit$coefficients),"\n")
				// }
            // } else if (method[1] == "svm") {

                // #dat.tmp=dat.train[dat.train$WW!=0,]
                // #dat.tmp$w.tmp=dat.train[dat.train$WW!=0,]$WW
                // #fit.svml=try.error(svml(formula,data=dat.tmp,weights=w.tmp,cost=1/(param$lamcost*sum(abs(dat.tmp$WW))),kernel="linear(l)",class.type='oao'))

                // fit.svml = try.error(kernlab.SVW1 (formula, dat=dat.train, W=dat.train$WW, s=1, lambda=param$lamcost*nrow(dat.train),
                    // kernel=kernel, para=param$gamma, verbose=F,beta.init=beta.init,arg=0))
                // #fit.svml=svml(formula,data=dat[dat$WW!=0,],weights=dat[dat$WW!=0,]$WW,cost=1/(param$lamcost*sum(abs(dat$WW))),kernel="linear(l)",class.type
                // #='oao')
                // pred.svml<-predict.svml2(fit.svml, newdata=dat.tune, case.percentage = NULL)
                // rts.tune=caltheta(dat.tune$yy,dat.tune$TT,pred.svml)
            // }else{
                // fit = try.error(rtswt(formula, dat=dat.train, weights=dat.train[,varW], s=1, lambda=param$lamcost*nrow(dat.train),
                    // kernel=kernel, para=param$gamma, verbose=F,beta.init=beta.init,tol = 1e-6))
				// fit$alpha.pred <- fit$alpha
				// fit$b0 <- fit$coefficients[1]
				// fit$coefficients <- fit$coefficients[-1]
                // pred.rts<-predict.rtw2(fit, newdata=dat.tune, case.percentage = NULL)
                // rts.tune=caltheta(dat.tune$yy,dat.tune$TT,pred.rts)
			
			// }
            // res = c(lamcost = param$lamcost, gamma = unname(param$gamma),tune = rts.tune)
            // if (verbose){
				// print(res)
			// }
            // res
        // })
    // # if (arg.tune==0) tmp = res[, which.min(res["tune", ])]
    // # else if (arg.tune==1) tmp = res[, which.min.random(res["tune", ])]
    // # else if (arg.tune==2) {
        // # if (kernel == "l")
        // # tmp = res[, which.min.mix(res[c("lamcost","tune"), ])]
        // # if (kernel == "r")
        // # tmp = res[, which.min.mix(res[c("lamcost","gamma","tune"), ])]
    // # }
    // # if (kernel == "r") {
        // # res = tmp[c("lamcost", "gamma","tune")]
    // # } else if (kernel == "l") {
        // # res = tmp[c("lamcost","tune")]
    // # }
    // # else stop("kernel not supported: " %+% kernel)
    // return(res)
// }
