// TODO: check rtsw(),rtswt() arguments to make it clear which arguments are i/o
// rtsw,rtswt,rts_svm - pass alpha from previous iteration as 'primal' starting point to loqo
// for this you may need to allocate space for alpha in rts functions

#include "rts.h"
#include "matrix.h"
#include <R.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#define	MAX(A,B)	((A) > (B) ? (A) : (B))
#define	MIN(A,B)	((A) < (B) ? (A) : (B))
#define	ABS(x)  	((x) > 0 ? (x) : (-(x)))
#define	SIGN(x)	    ((x) > 0 ? 1 : (x)==0 ? 0 : -1) 
#define EQ(x,y,eps) ((ABS((x)-(y))) <= (eps) ? 1 : 0) 

			 
// eta  : n x 1
// s    : 1x1 
// loss : logical T/F
// y    : n x 1 contains the result, may be the same as eta
void rampfun(int* n,double* eta,double* s,int* loss,double* y){
	if(*s == DBL_MAX){
		for(int i = 0;i < *n;i++)
			if(eta[i] < 0.0) y[i] = 1.0;else y[i] = 0.0;
	}else{
		for(int i = 0;i < *n;i++){
			double x = *s * eta[i];
			if((-0.5 < x) && (x < 0.5)) y[i] = 0.5 - x;
			else if(0.5 <= x) y[i] = 0.0;
			else y[i] = 1.0; 
		}
	}
	if(!(*loss))for(int i = 0;i < *n;i++)y[i] = 1.0 - y[i];
}

/*
// INPUT: 
nrx           : row-dimension of 'x',length of 'y','w'
ncx           : column-dim of 'x' without the intercept
y             : outcome 0/1
x             : matrix of covariates (column-major order)
w             : vector of 'weights' (-inf,inf)
is_intercept  : 0/1 whether to fit with intercept
is_eta        : should I use values in 'eta' for initialization
eta_init      : length nrx to initialize eta
is_beta       : should I use values in 'beta' for initialization
beta_init     : length 'ncx' (b_0, b_1,..,b_p)
s             : tuning parameter
lambda        : tuning parameter
kernel        : linear,...
kernel  : parameter to kernel
maxit         : i/o
tol           : i/o
verbose       :

OUTPUT:
convergence   : 0 if converged
beta          : (beta_0,beta_1..beta_p)
alpha_pred    : 
linear_combinations: 
eta           : eta from last iteration 
*/


// internal_buffer_size = n*n + 9*n + (p + 1)
// Q         n x n
// Qa_pred   n
// alpha_pred n not necessary but cleaner
// b         n
// dvec_n    n
// Z         n
// Wabs      n
// lo        n
// dh2_deta  n
// eta_old   n
// beta_old  p+1

// internal_buffer_size >= (*nrx) * (*nrx) + 9 * (*nrx) + (p+1) otherwise allocated		
void rtsw(RTS_MODEL_FRAME* data,RTS_PARAM* param,KERNEL_PARAM* kernel,RTS_FIT* fit){
	int i;
	int n  = data->nrx;
	int p  = data->ncx;
	int ione = 1;	
	RTS_CONTROL* control = &(param->control);	
	LOQO_CONTROL* lqcontrol = &(param->loqo_control); 
	LOQO_ARG loqo_arg;
	LOQO_ARG* lqarg = &(loqo_arg);memset(lqarg,0,sizeof(LOQO_ARG));
	KERNEL_ARG karg;memset(&karg,0,sizeof(KERNEL_ARG));
	karg.x1 = data->X;karg.x2 = data->X;karg.nr1 = n;karg.nr2 = n;karg.nc1 = data->ncx;karg.nc2 = data->ncx;
	
	double* buffer = NULL;
	int internal_buffer_size = n*n + 9*n + (p + 1);
	int internal_buffer = 1;
	if(param->aux_buffer && param->aux_buffer_size)
		internal_buffer = internal_buffer_size > param->aux_buffer_size;
	if(internal_buffer){
		buffer = (double*)calloc((size_t)internal_buffer_size , sizeof(double));  
	}else{
		buffer = param->aux_buffer;
		memset(buffer,0,(size_t)internal_buffer_size * sizeof(double));
	}
	double* _p = buffer; //to be incremented as needed
	int loqo_buffer_size = param->aux_buffer_size - internal_buffer_size; // leftover from external_buffer available to LOQO
	double* loqo_buffer = NULL;
	if(loqo_buffer_size) loqo_buffer = buffer + internal_buffer_size;   // LOQO will decide if this is enough

	double lambda = param->lambda;
	double s = param->s;
	double* X = data->X;
	double* y = data->y;
	double* w = data->w;
	double* alpha_pred = fit->alpha_pred;
	double* eta_new = fit->eta;
	double* beta_new = fit->beta;	
	// rtsw	
	double* Q = _p;_p+=n*n;
	double* Qa_pred =_p;_p+=n;
	double* b =_p;_p+=n;
	double* dvec_n =_p;_p+=n;
	double* Z =_p;_p+=n;
	double* Wabs =_p;_p+=n;
	double* dh2_deta = _p;_p+=n;
	double* eta_old =_p;_p+=n;
	double* beta_old =_p;_p+=(p+1);	


// LOQO
	double* primal =_p;_p+=n;
    double* lo = _p;_p+=n; // 0 <= x
    double* up = Wabs;	   // x <= Wabs
	double dual = lqcontrol->inf;
	double constr_vec1 = 0.0; // not an array since num_constr = 1
	double constr_vec2 = 0.0; // not an array since num_constr = 1
	lqarg->n = n;
	lqarg->m = 1; // number of constraints
	lqarg->H = Q;
	lqarg->b = &constr_vec1;
	lqarg->r = &constr_vec2;
	lqarg->A = Z;
	lqarg->primal = primal;
	lqarg->lo = lo;
	lqarg->up = up;
	lqarg->double_buffer_size = loqo_buffer_size;
	lqarg->double_buffer = loqo_buffer;	
	lqarg->dual = &dual;
	lqarg->primal_obj = lqcontrol->inf;
	lqarg->dual_obj = 0.5*lqcontrol->inf;	
	
	int n0 = 0;
	int n1 = 0;
	for(i = 0;i < n;i++) if(y[i] == 1.0) n1++;else n0++;
	for(i = 0;i < n;i++) Z[i] = SIGN(w[i]);
    for(i = 0;i < n;i++) Wabs[i] = ABS(w[i]);

	// int kernel = (int)fit->kernel;getKernel(&n,&p,X,&n,&p,X,&kernel,&fit->kernel,Q);
	getK(&karg,kernel,Q);
	for(i = 0;i < (n*n);i++)Q[i] /= lambda;    	
	dxd(&n,Z,Q,Z,Q); 
	

    if(param->beta_init) memcpy(beta_old,param->beta_init,(p + 1) * sizeof(double));
	
	// treat intercept separately as x does not contain intercept
	if(!(param->eta_init)) {
		if(kernel->type == LINEAR){
			//eta.init <- Z * (X %*% beta.init[-1])
			matprod(X,&n,&p,beta_old + 1,&p,&ione,eta_old);
			for(i = 0;i < n;i++) eta_old[i] *= Z[i];
			if(param->is_intercept)
				for(i = 0;i < n;i++) 
					eta_old[i] += Z[i] * beta_old[0];
		} else{
			; //eta_old is already set to 0 by calloc;
		}
	}else{
		memcpy(eta_old,param->eta_init,(size_t)n * sizeof(double));
	}
	
	
    //val.init <- sum(Wabs * ramp.f(eta.init,s)) + lambda*sum(beta.init[-1]^2)/2	
	double* ramp_eta_old = dvec_n;int loss = 0;
	rampfun(&n,eta_old,&s,&loss,ramp_eta_old);
	double val_old = 0.0;for(i = 0;i < n;i++) val_old += Wabs[i] * ramp_eta_old[i];
	if(lambda != 0.0){
		double pen = 0.0;
		for(int j = 0;j < p;j++)pen += beta_old[j+1] * beta_old[j+1];
		pen *= 0.5 * lambda;
		val_old += pen;		
	}	
	
	double b0_pred = 0.0;	
	
	
    ////////////////////////////////////////////////////////////////////////////////////////
	int iter = 0;
    fit->convergence = 1;
	while(1){
		iter++;
				
        //dh2.deta <- ifelse(eta.init < -0.5/s, -s, 0)
		for(i = 0;i < n;i++) if(eta_old[i] < -0.5/s) dh2_deta[i] = -s;else dh2_deta[i] = 0.0;
		
        // b <- s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))
		for(i = 0;i < n;i++) dvec_n[i] = Wabs[i] * dh2_deta[i];
		matprod(Q,&n,&n,dvec_n,&n,&ione,b);
		double s2 = (s * s);
		for(i = 0;i < n;i++) b[i] = (1.0 - 2.0 * s * b[i]) / s2;
		
		// fit <- loqo(as.matrix(-0.5*b),Q,t(as.matrix(Z)),b = -sum(W * dh2.deta),l = rep(0,n),u=Wabs,r=0)
		double* c = dvec_n;for(i = 0;i < n;i++) c[i] = -0.5 * b[i]; //_b = -0.5*b
		lqarg->c = c;
		lqarg->b[0] = 0.0;for(i = 0;i < n;i++) lqarg->b[0] -= w[i] * dh2_deta[i];
		
		loqo(lqarg,lqcontrol);		
	
		if((control->verbose > 0) && lqarg->status){
			Rprintf("LOQO did not converge at iteration %i, returned with status %i: %s, error code %i\n",iter,
				lqarg->status,LOQO_STATUS_NAMES[lqarg->status],lqarg->error_code);
		}
		
		// // allow iteration to continue if convergence is slow (maxit is reached) and let rts convergence decide
		// if((lqarg->status != LOQO_CONVERGED) && (lqarg->status != LOQO_SLOW_CONVERGENCE)){
			// // Rprintf("(lo <= x <= up\n");
			// // for(i = 0;i < n;i++) Rprintf("%.3e <= %.3e <= %.3e\n",lo[i],primal[i],up[i]);
			// // Rprintf("\n");
			// break;
		// }


        // alpha.pred = dh2.deta * Wabs + s * fit$alpha
		// Qa.pred <- Q %*% alpha.pred
        // pen = 0.5*drop(crossprod(alpha.pred, Qa.pred))
		for(i = 0;i < n;i++) alpha_pred[i] = Wabs[i] * dh2_deta[i] + s * primal[i];
		matprod(Q,&n,&n,alpha_pred,&n,&ione,Qa_pred);		
		double pen;crossprod(alpha_pred,&n,&ione,Qa_pred,&n,&ione,&pen); 
		pen *= 0.5;
		// Rprintf("Qa_pred ");for(i = 0;i < n;i++)Rprintf("%.4e ",Qa_pred[i]);Rprintf("\n");

		
        ////////// typical support vector machine way to get intercept b0
		// b0.pred <- 0
        // if (INTERCEPT){
			// if(fit$nFSV > 0){
				// b0.pred = 1/Z * (1/2 - Qa.pred)
				// b0.pred = mean(b0.pred[fit$alpha>0 & fit$alpha < Wabs])
			// }
        // }
		// fit$nSV  <- sum(0.0 < fit$alpha)
		// fit$nBSV <- sum(fit$alpha == Wabs)
		// fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < Wabs))
		
		b0_pred = 0.0;
		int nSV = 0;
		int nBSV = 0;
		int nFSV = 0;
		for(i = 0;i < n;i++){
			int left = (int)(lo[i] < primal[i]);	
			int right = (int)(primal[i] < up[i]);	
			nSV  += left;
			nBSV += !right;
			int isFSV = left && right;
			nFSV += isFSV; 
			if(isFSV && (param->is_intercept)) b0_pred += (0.5 - Qa_pred[i]) / Z[i];			
		}			
		if(nFSV && param->is_intercept)b0_pred /= (double)nFSV;
		beta_new[0] = b0_pred;
		// Rprintf("b0=%.4f beta_new[0]=%.4f\n",b0_pred,beta_new[0]);
		
		
		// eta.new = Z * b0.pred + Qa.pred
        // val.new = sum(Wabs * ramp.f(eta.new,s))
		for(i = 0;i < n;i++) eta_new[i] = b0_pred * Z[i] + Qa_pred[i];
		double* ramp_eta_new = dvec_n;int loss = 0;
		rampfun(&n,eta_new,&s,&loss,ramp_eta_new);
        double val_new = 0.0;for(i = 0;i < n;i++) val_new += Wabs[i] * ramp_eta_new[i];
		// Rprintf("eta_new ");for(i = 0;i < n;i++)Rprintf("%.4e ",eta_new[i]);Rprintf("\n");
					
       // check stopping criteria and udpate eta.init
		double delta = DBL_MAX;
        // if ((KERNEL_TYPE)(*kernel) == LINEAR){
			// for(i = 0;i < n;i++)  dvec_n[i] = alpha_pred[i] * Z[i]; 
            // crossprod(X,&n,&p,dvec_n,&n,&ione,beta_new + 1);
			// for(i = 1;i < (p+1);i++) beta_new[i] /= lambda;

		// // delta=max(abs((beta.new[-1]-beta.init[-1])/beta.init[-1])) # skip intercept			
			// delta = ABS((beta_new[1]-beta_old[1]) / beta_old[1]);
			// for(i = 1;i < p;i++) delta = MAX(delta,ABS((beta_new[i+1]-beta_old[i+1]) / beta_old[i+1]));

			// if(verbose > 0){
				// Rprintf(", beta");
				// for(i = 0;i < p+1;i++)Rprintf(" %.4f ",beta_new[i]);
				// Rprintf("delta %.4f\n",delta);
			// }
            // if(delta < tol) {
                // *convergence = 0;
                // break;
            // } else {
				// memcpy(beta_old,beta_new,(p + 1) * sizeof(double));
            // }
        // } else 
		{
            val_new += pen;
            delta = val_new - val_old;
            if(control->verbose > 1){ 
				double saturation = 0.0;
				for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
				saturation /= (double)n;
				Rprintf("%i '%s' saturated %.3e loss %.3e pen %.3e delta %.3e\n",
						iter,LOQO_STATUS_NAMES[lqarg->status],saturation,val_new,pen,delta);			
			}
            if (ABS(delta) < control->tol) {
                fit->convergence = 0;
                break;
            } else {
                val_old = val_new;
            }
        }	
		memcpy(eta_old,eta_new,n * sizeof(double));		
		if(iter == control->maxit)break;
	}	
	if (control->verbose > 0) Rprintf("\n");			

    // Z.pred = Z
    // Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda

	double* Q_pred = Q;
	double* Z_pred = Z;		
  		
    // linear.combination = Z.pred * b0.pred + Q.pred %*% alpha.pred
	double* linear_combination = fit->linear_combination;
	matprod(Q_pred,&n,&n,alpha_pred,&n,&ione,linear_combination);
	for(i = 0;i < n;i++)
		linear_combination[i] += Z_pred[i] * b0_pred;
	for(i = 0;i < n;i++)
		if(EQ(linear_combination[i],0.0,DBL_EPSILON))
			linear_combination[i] = 0.0;			

	if (kernel->type == LINEAR){
		double* beta_new = fit->beta;
		for(i = 0;i  < n;i++)  dvec_n[i] = alpha_pred[i] * Z[i]; 
		crossprod(X,&n,&p,dvec_n,&n,&ione,beta_new + 1);
		for(i = 1;i < (p+1);i++) beta_new[i] /= lambda;
	}

	if(control->verbose == 1){ 
		double saturation = 0.0;
		for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
		saturation /= (double)n;
		Rprintf("%i '%s' saturated \n",iter,LOQO_STATUS_NAMES[lqarg->status],saturation);			
	}

	// these are aliases only
	memcpy(&fit->param,param,sizeof(RTS_PARAM));				
	memcpy(&fit->kernel,kernel,sizeof(KERNEL_PARAM));
	memcpy(&fit->data,data,sizeof(RTS_MODEL_FRAME));				
	
	if(internal_buffer)free(buffer);
}
void R_rtsw(
// input data
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta
){
	RTS_FIT Fit;memset(&Fit,0,sizeof(RTS_FIT));
	KERNEL_PARAM Kernel;memset(&Kernel,0,sizeof(KERNEL_PARAM));	
	RTS_MODEL_FRAME Data;memset(&Data,0,sizeof(RTS_MODEL_FRAME));
	RTS_MODEL_FRAME* data = &Data; 
	RTS_PARAM Param;memset(&Param,0,sizeof(RTS_PARAM));
	RTS_CONTROL* control = &(Param.control);	
	LOQO_CONTROL* c = &Param.loqo_control;
	*convergence = 1; //bad

// control arguments should arrive in alphabetical order - dangerous practice	
	
	// rts control
	control->maxit = (int)rts_control[0];
	control->tol = rts_control[1];
	control->verbose = (int)rts_control[2];
	
	// quadratic optimizer control
	c->bound = loqo_control[0]; 
	c->inf = loqo_control[1];
	c->margin = loqo_control[2]; 
	c->maxiter = (int)loqo_control[3]; 
	c->sigfig_max = loqo_control[4];
	c->lapack_solver = (LAPACK_SOLVER)loqo_control[5];
	c->verbose = (int)loqo_control[6];	
	
	data->nrx = *nrx;
	data->ncx = *ncx;
	data->X   = X;
	data->y = y;
	data->w = w;
	
	Param.s = *s;
	Param.lambda = *lambda;
	// fit.threshold = *threshold;
	
	Param.is_intercept = *is_intercept;
	if(*is_eta){
		Param.neta = *nrx;
		Param.eta_init = eta_init;
	}
	if(*is_beta){
		Param.nbeta = *ncx + 1;
		Param.beta_init = beta_init;
	}
	Kernel.type = (KERNEL_TYPE)*kernel;
	Kernel.param = *kernel_param;
	
	Fit.beta = beta;
	Fit.alpha_pred = alpha_pred;
	Fit.linear_combination = linear_combination;
	Fit.eta = eta;

	rtswt(&Data,&Param,&Kernel,&Fit);
	
	*convergence = Fit.convergence;	
}


// Implements kernlab.W1
// It does not take the arguments "start.method" and "fit"
// internal_buffer_size >= (*nrx) * (*nrx) + 9 * (*nrx) + (*ncx + 1) otherwise allocated
void rtswt(RTS_MODEL_FRAME* data,RTS_PARAM* param,KERNEL_PARAM* kernel,RTS_FIT* fit){

	int i;
	int n  = data->nrx;
	int p  = data->ncx;
	int ione = 1;	
	RTS_CONTROL* control = &(param->control);	
	LOQO_CONTROL* lqcontrol = &(param->loqo_control); 
	LOQO_ARG loqo_arg;
	LOQO_ARG* lqarg = &(loqo_arg);memset(lqarg,0,sizeof(LOQO_ARG));
	KERNEL_ARG karg;memset(&karg,0,sizeof(KERNEL_ARG));
	karg.x1 = data->X;karg.x2 = data->X;karg.nr1 = n;karg.nr2 = n;karg.nc1 = data->ncx;karg.nc2 = data->ncx;
	
	// Rprintf("\n#######################################################################\n");
	// print_control_rts(control);
	// print_param_rts(param);
	// print_mf_rts(data,0);
	// Rprintf("#######################################################################\n\n");
		
		
	double* buffer = NULL;
	int internal_buffer_size = n*n + 9*n + (p + 1);
	int internal_buffer = 1;
	if(param->aux_buffer && param->aux_buffer_size)
		internal_buffer = internal_buffer_size > param->aux_buffer_size;
	if(internal_buffer){
		buffer = (double*)calloc((size_t)internal_buffer_size , sizeof(double));  
	}else{
		buffer = param->aux_buffer;
		memset(buffer,0,(size_t)internal_buffer_size * sizeof(double));
	}
	double* _p = buffer; //to be incremented as needed
	int loqo_buffer_size = param->aux_buffer_size - internal_buffer_size; // leftover from external_buffer available to LOQO
	double* loqo_buffer = NULL;
	if(loqo_buffer_size) loqo_buffer = buffer + internal_buffer_size;   // LOQO will decide if this is enough

	double threshold = param->threshold;
	double lambda = param->lambda;
	double s = param->s;	
	double* X = data->X;
	double* y = data->y;
	double* w = data->w;
	double* alpha_pred = fit->alpha_pred;
	double* eta_new = fit->eta;
	double* beta_new = fit->beta;
	
	double* Q = _p;_p+=n*n;
	double* Qa_pred =_p;_p+=n;
	double* b =_p;_p+=n;
	double* dvec_n =_p;_p+=n;
	double* Z =_p;_p+=n;
	double* Wabs =_p;_p+=n;
	double* dh2_deta = _p;_p+=n;
	double* eta_old =_p;_p+=n;
	double* beta_old =_p;_p+=(p+1);



// LOQO control arguments should arrive in alphabetical order - dangerous practice	
	double* primal =_p;_p+=n;
    double* lo = _p;_p+=n; // 0 <= x
    double* up = Wabs;	   // x <= Wabs
	double dual = lqcontrol->inf;
	double constr_vec1 = 0.0; // not an array since num_constr = 1
	double constr_vec2 = 0.0; // not an array since num_constr = 1
	lqarg->n = n;
	lqarg->m = 1; // number of constraints
	lqarg->H = Q;
	lqarg->b = &constr_vec1;
	lqarg->r = &constr_vec2;
	lqarg->A = Z;
	lqarg->primal = primal;
	lqarg->lo = lo;
	lqarg->up = up;
	lqarg->double_buffer_size = loqo_buffer_size;
	lqarg->double_buffer = loqo_buffer;	
	lqarg->dual = &dual;
	lqarg->primal_obj = lqcontrol->inf;
	lqarg->dual_obj = 0.5*lqcontrol->inf;	
		
	int n0 = 0;
	int n1 = 0;
	for(i = 0;i < n;i++) if(y[i] == 1.0) n1++;else n0++;
	for(i = 0;i < n;i++) Z[i] = SIGN(w[i]);
    for(i = 0;i < n;i++) Wabs[i] = ABS(w[i]);


	// int kernel = (int)fit->kernel;getKernel(&n,&p,X,&n,&p,X,&kernel,&fit->kernel,Q);	
	getK(&karg,kernel,Q);
	for(i = 0;i < (n*n);i++)Q[i] /= lambda;    	
	dxd(&n,Z,Q,Z,Q); 
	

    if(param->beta_init) memcpy(beta_old,param->beta_init,(p + 1) * sizeof(double));

	// treat intercept separately as x does not contain intercept
	if(!(param->eta_init)) {
		if(kernel->type == LINEAR){
			//eta.init <- Z * (X %*% beta.init[-1])
			matprod(X,&n,&p,beta_old + 1,&p,&ione,eta_old);
			for(i = 0;i < n;i++) eta_old[i] *= Z[i];
			if(param->is_intercept)
				for(i = 0;i < n;i++) 
					eta_old[i] += Z[i] * beta_old[0];
		} else{
			; //eta_old is already set to 0 by calloc;
		}
	}else{
		memcpy(eta_old,param->eta_init,(size_t)n * sizeof(double));
	}
	
	
    //val.init <- sum(Wabs * ramp.f(eta.init,s)) + lambda*sum(beta.init[-1]^2)/2	
	double* ramp_eta_old = dvec_n;int loss = 0;
	rampfun(&n,eta_old,&s,&loss,ramp_eta_old);
	double val_old = 0.0;for(i = 0;i < n;i++) val_old += Wabs[i] * ramp_eta_old[i];
	if(lambda != 0.0){
		double pen = 0.0;
		for(int j = 0;j < p;j++)pen += beta_old[j+1] * beta_old[j+1];
		pen *= 0.5 * lambda;
		val_old += pen;		
	}
	
	double b0_pred = 0.0;	
	
    ////////////////////////////////////////////////////////////////////////////////////////
	int iter = 0;
    fit->convergence = 1;
	while(1){
		iter++;
				
        //dh2.deta <- ifelse(eta.init < -0.5/s, -s, 0)
		for(i = 0;i < n;i++) if(eta_old[i] < -0.5/s) dh2_deta[i] = -s;else dh2_deta[i] = 0.0;
		
        // b <- s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))
		for(i = 0;i < n;i++) dvec_n[i] = Wabs[i] * dh2_deta[i];
		matprod(Q,&n,&n,dvec_n,&n,&ione,b);
		double s2 = (s * s);
		for(i = 0;i < n;i++) b[i] = (1.0 - 2.0 * s * b[i]) / s2;
		
		// fit <- loqo(as.matrix(-0.5*b),Q,t(as.matrix(Z)),b = -sum(W * dh2.deta),l = rep(0,n),u=Wabs,r=0)
		double* c = dvec_n;for(i = 0;i < n;i++) c[i] = -0.5 * b[i]; //_b = -0.5*b
		lqarg->c = c;
		lqarg->b[0] = 0.0;for(i = 0;i < n;i++) lqarg->b[0] -= w[i] * dh2_deta[i];
		
		loqo(lqarg,lqcontrol);		
			
		if((control->verbose > 0) && lqarg->status){
			Rprintf("LOQO did not converge at iteration %i, returned with status %i: %s, error code %i\n",iter,
				lqarg->status,LOQO_STATUS_NAMES[lqarg->status],lqarg->error_code);
		}
		
		// // allow iteration to continue if convergence is slow (maxit is reached) and let rts convergence decide
		// if((lqarg->status != LOQO_CONVERGED) && (lqarg->status != LOQO_SLOW_CONVERGENCE)){
			// // Rprintf("(lo <= x <= up\n");
			// // for(i = 0;i < n;i++) Rprintf("%.3e <= %.3e <= %.3e\n",lo[i],primal[i],up[i]);
			// // Rprintf("\n");
			// break;
		// }


        //alpha.pred = dh2.deta * Wabs + s * fit$alpha
		for(i = 0;i < n;i++) alpha_pred[i] = Wabs[i] * dh2_deta[i] + s * primal[i];
	
		//Qa.pred <- Q %*% alpha.pred
		matprod(Q,&n,&n,alpha_pred,&n,&ione,Qa_pred);
		double pen;crossprod(alpha_pred,&n,&ione,Qa_pred,&n,&ione,&pen); 
		pen *= 0.5;
			
        ////////// typical support vector machine way to get intercept b0
		// b0.pred <- 0
        // if (INTERCEPT){
			// if(fit$nFSV > 0){
				// b0.pred = 1/Z * (1/2 - Qa.pred)
				// b0.pred = mean(b0.pred[fit$alpha>0 & fit$alpha < Wabs])
			// }
        // }
		// fit$nSV  <- sum(0.0 < fit$alpha)
		// fit$nBSV <- sum(fit$alpha == Wabs)
		// fit$nFSV <- sum((0.0 < fit$alpha) & (fit$alpha < Wabs))
		
		b0_pred = 0.0;
		int nSV = 0;
		int nBSV = 0;
		int nFSV = 0;
		double* alpha = primal;
		for(i = 0;i < n;i++)alpha[i] /= Wabs[i]; // primal[i] [lo[i],up[i] -> [0,1]
		
		for(i = 0;i < n;i++){
			int left = (int)(threshold < alpha[i]);	
			int right = (int)(alpha[i] < 1.0 - threshold);	
			nSV  += left;
			nBSV += !right;
			int isFSV = left && right;
			nFSV += isFSV; 
			if(isFSV && param->is_intercept) b0_pred += (0.5 - Qa_pred[i]) / Z[i];			
		}			
		
		if(param->is_intercept){
			if(nFSV >= 2){
				b0_pred /= (double)nFSV;
			}else{
				b0_pred = 0.0;
				
				// top 5 closest to thresholded-bound
				// top_k_index <- order(apply(cbind(alpha,1.0-alpha),1,min),decreasing=TRUE)[1:min(5,n)] 			
				// top_k_alpha will hold the top 5 a's instead of the indeces (in alpha) of the top 5 a's				
				double top_k_alpha[MIN(5,n)]; 
				int top_k_index[MIN(5,n)];
				for(int j = 0;j < MIN(5,n);j++)top_k_alpha[j] = -DBL_MAX;
				for(i = 0;i < n;i++){
					double a = MIN(alpha[i],1.0-alpha[i]);					
					int ax = i;
					for(int j = 0;j < MIN(5,n);j++){
						if(a > top_k_alpha[j]){
							double tmp = top_k_alpha[j];
							top_k_alpha[j] = a;
							a = tmp;
							int tx = top_k_index[j];
							top_k_index[j] = ax;
							ax = tx;
						}
					}			
				}
				if(control->verbose > 1){
					Rprintf("threshold free alpha:");
					for(int j = 0;j < MIN(5,n);j++)Rprintf("%.6f ",top_k_alpha[j]);Rprintf("\n");
				}
				for(int j = 0;j < MIN(5,n);j++)		
					b0_pred +=  (0.5 - Qa_pred[top_k_index[j]]) / Z[top_k_index[j]];
				b0_pred /= (double)(MIN(5,n));
			}
		}
		
		beta_new[0] = b0_pred;
		
		// if(control->verbose > 1)Rprintf("b0.pred=%.10f\n",b0_pred);
		// Rprintf("Qa_pred ");for(i = 0;i < n;i++)Rprintf("%.4e ",Qa_pred[i]);Rprintf("\n");
		
		// eta.new = Z * b0.pred + Qa.pred
        // val.new = sum(Wabs * ramp.f(eta.new,s))
		for(i = 0;i < n;i++) eta_new[i] = b0_pred * Z[i] + Qa_pred[i];
		// Rprintf("eta_new ");for(i = 0;i < n;i++)Rprintf("%.4e ",eta_new[i]);Rprintf("\n");

		double* ramp_eta_new = dvec_n;int loss = 0;
		rampfun(&n,eta_new,&s,&loss,ramp_eta_new);
        double val_new = 0.0;for(i = 0;i < n;i++) val_new += Wabs[i] * ramp_eta_new[i];
				
		// if (verbose > 1) cat(", saturated ", signif(mean(eta.new > 0.5/s | eta.new < -0.5/s), 3))
			    	
       // check stopping criteria and udpate eta.init
		double delta = DBL_MAX;
        // if ((KERNEL_TYPE)(*kernel) == LINEAR){
			// for(i = 0;i < n;i++)  dvec_n[i] = alpha_pred[i] * Z[i]; 
            // crossprod(X,&n,&p,dvec_n,&n,&ione,beta_new + 1);
			// for(i = 0;i < p;i++) beta_new[i+1] /= lambda;

            // // delta=max(abs((beta.new[-1]-beta.init[-1])/beta.init[-1])) # skip intercept			
			// delta = ABS((beta_new[1]-beta_old[1]) / beta_old[1]);
			// for(i = 1;i < p;i++) delta = MAX(delta,ABS((beta_new[i+1]-beta_old[i+1]) / beta_old[i+1]));

			// if(verbose > 0){Rprintf(", beta");for(i = 0;i < p+1;i++)Rprintf(" %.4f ",beta_new[i]);Rprintf("delta %.4f\n",delta);}
            // if(delta < tol) {
                // *convergence = 0;
                // break;
            // } else {
				// memcpy(beta_old,beta_new,(p + 1) * sizeof(double));
            // }
        // } else 
		{
            val_new += pen;
            delta = val_new - val_old;
            if(control->verbose > 1){ 
				double saturation = 0.0;
				for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
				saturation /= (double)n;
				Rprintf("%i '%s' saturated %.3e loss %.3e pen %.3e delta %.3e\n",
						iter,LOQO_STATUS_NAMES[lqarg->status],saturation,val_new,pen,delta);			
			}
            if (ABS(delta) < control->tol) {
                fit->convergence = 0;
                break;
            } else {
                val_old = val_new;
            }
        }	
		memcpy(eta_old,eta_new,n * sizeof(double));		
		if(iter == control->maxit)break;
	}	
	if (control->verbose > 0) Rprintf("\n");			

    // Z.pred = Z
    // Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
	double* Q_pred = Q;
	double* Z_pred = Z;
	// getKernel(&n,&p,X,&n,&p,X,kernel,kernel,Q_pred);	
	// dxd(&n,Z_pred,Q_pred,Z_pred,Q_pred); 
	// for(i = 0;i < n*n;i++)Q_pred[i] /= lambda;    
		
    // linear.combination = Z.pred * b0.pred + Q.pred %*% alpha.pred
	double* linear_combination = fit->linear_combination;
	matprod(Q_pred,&n,&n,alpha_pred,&n,&ione,linear_combination);
	for(i = 0;i < n;i++)
		linear_combination[i] += Z_pred[i] * b0_pred;
	for(i = 0;i < n;i++)
		if(EQ(linear_combination[i],0.0,DBL_EPSILON))
			linear_combination[i] = 0.0;			

	if (kernel->type == LINEAR){
		double* beta_new = fit->beta;
		for(i = 0;i  < n;i++)  dvec_n[i] = alpha_pred[i] * Z[i]; 
		crossprod(X,&n,&p,dvec_n,&n,&ione,beta_new + 1);
		for(i = 1;i < (p+1);i++) beta_new[i] /= lambda;
	}

	if(control->verbose == 1){ 
		double saturation = 0.0;
		for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
		saturation /= (double)n;
		Rprintf("%i '%s' saturated \n",iter,LOQO_STATUS_NAMES[lqarg->status],saturation);			
	}
			
	// these are aliases only
	memcpy(&fit->param,param,sizeof(RTS_PARAM));				
	memcpy(&fit->kernel,kernel,sizeof(KERNEL_PARAM));
	memcpy(&fit->data,data,sizeof(RTS_MODEL_FRAME));							
	if(internal_buffer)free(buffer);
}

void R_rtswt(
// input data
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta
){
	RTS_FIT Fit;memset(&Fit,0,sizeof(RTS_FIT));
	KERNEL_PARAM Kernel;memset(&Kernel,0,sizeof(KERNEL_PARAM));	
	RTS_MODEL_FRAME Data;memset(&Data,0,sizeof(RTS_MODEL_FRAME));
	RTS_MODEL_FRAME* data = &Data; 
	RTS_PARAM Param;memset(&Param,0,sizeof(RTS_PARAM));
	RTS_CONTROL* control = &(Param.control);	
	LOQO_CONTROL* c = &Param.loqo_control;
	*convergence = 1; //bad
	
// control arguments should arrive in alphabetical order - dangerous practice		
	// rts control
	control->maxit = (int)rts_control[0];
	control->tol = rts_control[1];
	control->verbose = (int)rts_control[2];
	
	// quadratic optimizer control
	c->bound = loqo_control[0]; 
	c->inf = loqo_control[1];
	c->margin = loqo_control[2]; 
	c->maxiter = (int)loqo_control[3]; 
	c->sigfig_max = loqo_control[4];
	c->lapack_solver = (LAPACK_SOLVER)loqo_control[5];
	c->verbose = (int)loqo_control[6];	
	
	data->nrx = *nrx;
	data->ncx = *ncx;
	data->X   = X;
	data->y = y;
	data->w = w;
		
	Param.s = *s;
	Param.lambda = *lambda;
	Param.threshold = *threshold;
	
	Param.is_intercept = *is_intercept;
	if(*is_eta){
		Param.neta = *nrx;
		Param.eta_init = eta_init;
	}
	if(*is_beta){
		Param.nbeta = *ncx + 1;
		Param.beta_init = beta_init;
	}
	

	Kernel.type = (KERNEL_TYPE)*kernel;
	Kernel.param = *kernel_param;
	
	Param.aux_buffer_size = *aux_buffer_size;
	if(Param.aux_buffer_size)Param.aux_buffer = aux_buffer;
	
	Fit.beta = beta;
	Fit.alpha_pred = alpha_pred;
	Fit.linear_combination = linear_combination;
	Fit.eta = eta;
		
	rtswt(&Data,&Param,&Kernel,&Fit);
	
	*convergence = Fit.convergence;	
}



// # The R function implemented above:
// kernlab.W1=function (formula, dat, W, s = 1, lambda = 1, kernel = "linear",
    // para = NULL, start.method = "rlogit", eta0.init = NULL,threshold = 1e-4,
    // maxit = 100, tol = 1e-03, verbose = F, minQuad.tol = NULL, cost=0, beta.init,arg=0)
// {

    // b0.seq=seq(-2,2,0.01)
    // time0 = Sys.time()
    // tmp = model.frame(formula, dat)
    // X = model.matrix(formula, tmp)[, -1, drop = F]  ###
    // #y = dat$y
    // n = nrow(X)  ###
    // TT = dat$TT


    // ### get rid of points where W=0
    // #W<--(y*TT/sum(TT==1)-y*(1-TT)/sum(TT==0)+cost/n)
    // oo<-W!=0
    // W<-W[oo]
    // X<-as.matrix(X[oo,])
    // #y<-y[oo]




    // Z<-sign(W)
    // Wabs<-abs(W)   # absolute value for W
    // Wm<-diag(Wabs)
    // Zm<-diag(Z)
    // #X.diff = get.X.diff(X1, X2)
    // kernel = substr(kernel, 1, 1)
    // if (kernel != "l" & is.null(para))
        // stop("kernel parameter not set")
    // #X = rbind(X1, X2)
    // K = getK(X, kernel, para)/lambda
    // #Q = matrix(0, n1 * n2, n1 * n2)
    // Q = matrix(0, nrow(X), nrow(X))
    // if (!is.double(K))
        // K <- as.double(K)
	
    // Q = DXD(Zm, K, t(Zm))
// #    Q = crossprod(Zm, tcrossprod(K,Zm))



    // if (is.null(eta0.init)) {

        // if (kernel == "l"){
          // eta.init = Z * (cbind(1,X) %*% beta.init)
        // } else
        // eta.init=rep(0,sum(oo))
    // } else eta.init=eta0.init

   // val.init = sum(Wabs*ramp.f(eta.init,s))+lambda*sum(beta.init[-1]^2)/2

    // #print(val.approx(eta.init,Z,X,beta.init[1],beta.init[2],Wabs,lambda))
// #
// #    beta.seq<-c(beta.init[2],beta.new,beta.new1)
// #    val.seq<-val.approx(eta.init,Z,X,beta.init[1],beta.seq,Wabs,lambda)


    // iter = 0
    // converge = 0
    // while (TRUE) {
        // iter = iter + 1
        // if (iter > maxit) {
            // converge = 1
            // break
        // }
        // dh2.deta = ifelse(eta.init < -0.5/s, -s, 0)
        // b = s^(-2) * (1 - 2 * s * Q %*% (Wabs * dh2.deta))

        // fit<-ipop(as.matrix(-0.5*b),Q,t(as.matrix(Z)),b=-sum(W*dh2.deta),l=rep(0,length(Wabs)),u=Wabs,r=0)


        // if (how(fit) != 'converged')
            // print("minQuad does not converge in Iter " %+% iter)
        // if (verbose)
            // cat(iter)
        // pp<-primal(fit)
        // alpha.pred = dh2.deta * Wabs + s * pp

		// #cat("alpha_pred ",alpha_pred,"\n")
        // if (arg==0){
            // b0.pred = 1/Z * (1/2 - Q %*% alpha.pred)
            // oo.sp = pp/Wabs>threshold & pp/Wabs<1-threshold

            // if (sum(oo.sp)<2) {
                 // pp.min<-apply(cbind(pp/Wabs,1-pp/Wabs),1,min)
                 // oo.sp=order(pp.min,decreasing=TRUE)[1:5]
				 // cat("threshold free alphas ",(pp/Wabs)[oo.sp],"\n")				 
            // }

            // b0.pred = mean(b0.pred[oo.sp])
        // }

        // ##### directly minimizing the loss to get b0
        // if (arg==1){
			// eta.star=Q %*% alpha.pred

			// ETA.seq = tcrossprod(cbind(Z,eta.star) , cbind(b0.seq,1))
				// theta=colSums(as.vector(abs(W)) * rp.f(ETA.seq, s, loss=T))
				// oo.b0<-(1:length(b0.seq))[theta==min(theta)]
				// b0.pred=resample(b0.seq[oo.b0])[1]
        // }

        // #b0.pred=beta.init[1]
        // eta.new = Z * b0.pred + Q %*% alpha.pred
        // val.new = sum (Wabs * ramp.f(eta.new,s))
        // pen = drop(crossprod(alpha.pred, Q %*% alpha.pred))/2

        // if (verbose)
            // cat(", saturated ", signif(mean(eta.new > 0.5/s |
                // eta.new < -0.5/s), 3))
        // # if (kernel == "l") {
            // # beta.new = drop(crossprod(X, alpha.pred * Z)/lambda)
            // # beta.new = c(b0.pred, beta.new)
			
            // # if (verbose) cat(", beta[1] ", signif(beta.new[1], 4), ", beta[-1] ",
            // # signif(beta.new[-1], 4))

            // # if (verbose) {
                // # cat(", loss ", val.new, ", pen ", pen)
				// # val.new = val.new + pen
                // # cat(", val.new ", val.new)
            // # }

            // # delta = max(abs((beta.new - beta.init)/beta.init))
            // # if (delta < tol) {
                // # converge = 0
                // # break
            // # }
            // # else {
                // # beta.init = beta.new
                // # eta.init = eta.new
            // # }
        // # }
        // # else 
		// {
            // if (verbose)  cat(", loss ", val.new, ", pen ", pen)
            // val.new = val.new + pen
            // delta = val.new - val.init
            // if (abs(delta) < tol) {
                // converge = 0
                // break
            // }
            // else {
                // val.init = val.new
                // eta.init = eta.new
            // }
        // }
        // if (verbose)
            // cat("\n")
    // }
    // if (verbose)
        // cat("\n")

    // Z.pred = Z
    // Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
    // #linear.combination = Q.pred %*% alpha.pred

    // linear.combination = Z.pred * b0.pred + Q.pred %*% alpha.pred
    // result = list(converge = converge, iteration = iter, 
		// b0=b0.pred,beta = drop(beta.new[-1]),coefficients = drop(beta.new), 
		// W=W,X=X,
        // alpha.pred = drop(alpha.pred), linear.combination = drop(linear.combination),
        // last.minQuad.fit = fit,
        // kernel = kernel, para = para, formula = formula,lambda=lambda)
		
    // class(result) = c("rauc", class(result))
    // # if (kernel == "l") {
        // # result$coefficients = drop(t(t(alpha.pred) %*% X.diff)/lambda)
        // # result$coefficients = drop(crossprod(X, alpha.pred * Z)/lambda)
        // # result$coefficients = beta.new

    // # }
    // # if (result$converge != 0) {
        // # result$coefficients = rep(NA, ncol(X))
        // # result$alpha.pred = NA
        // # result$linear.combination = NA
    // # }
    // return(result)
// }
 
	
 
