#include "rts.h"
#include "loqo.h"
#include "matrix.h"
#include "kernels.h"
#include <R.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#define	MAX(A,B)	((A) > (B) ? (A) : (B))
#define	MIN(A,B)	((A) < (B) ? (A) : (B))
#define	ABS(x)  	((x) > 0 ? (x) : (-(x)))
#define	SIGN(x)	    ((x) > 0 ? 1 : (x)==0 ? 0 : -1) 
#define EQ(x,y,eps) ((ABS((x)-(y))) <= (eps) ? 1 : 0) 
#define HINGE(x)    ((x) > 0 ? (x) : ((0.0)))


// Implements the R function kernlab.SVW1 pasted at end of file
// It does not take the arguments "start.method" and "arg"
// internal_buffer_size >= (*nrx) * (*nrx) + 9 * (*nrx) + (*ncx + 1) otherwise allocated
void rtswt_svm(RTS_MODEL_FRAME* data,RTS_PARAM* param,KERNEL_PARAM* kernel,RTS_FIT* fit)
{

	int i;
	int n  = data->nrx;
	int p  = data->ncx;
	int ione = 1;			
	RTS_CONTROL* control = &(param->control);	
	LOQO_CONTROL* lqcontrol = &(param->loqo_control); 
	LOQO_ARG loqo_arg;
	LOQO_ARG* lqarg = &(loqo_arg);memset(lqarg,0,sizeof(LOQO_ARG));
	KERNEL_ARG karg;memset(&karg,0,sizeof(KERNEL_ARG));
	karg.x1 = data->X;karg.nr1 = n;karg.nc1 = data->ncx;
	karg.x2 = data->X;karg.nr2 = n;karg.nc2 = data->ncx;
		
	double* buffer = NULL;
	int internal_buffer_size = 2*n*n + 12*n + (p + 1);
	int internal_buffer = 1;
	if(param->aux_buffer && param->aux_buffer_size)
		internal_buffer = internal_buffer_size > param->aux_buffer_size;
	if(internal_buffer){
		buffer = (double*)calloc((size_t)internal_buffer_size , sizeof(double));  
	}else{
		buffer = param->aux_buffer;
		memset(buffer,0,(size_t)internal_buffer_size * sizeof(double));
	}
	double* _p = buffer; //to be incremented as needed
	int loqo_buffer_size = param->aux_buffer_size - internal_buffer_size; // leftover from external_buffer available to LOQO
	double* loqo_buffer = NULL;
	if(loqo_buffer_size) loqo_buffer = buffer + internal_buffer_size;   // LOQO will decide if this is enough
	
	double threshold = param->threshold;
	double lambda = param->lambda;
	double s = param->s;
	double* X = data->X;
	double* y = data->y;
	double* w = data->w;
	double* alpha_pred = fit->alpha_pred;
	double* eta_new = fit->eta;
	double* beta_new = fit->beta;
	
	double* H = _p;_p+=(n*n);
	double* Q = _p;_p+=(n*n);	
	double* Qa_pred =_p;_p+=n;
	double* Qya_pred =_p;_p+=n;
	double* y_alpha_pred = _p;_p+=n;
	double* alpha =_p;_p+=n;
	double* c =_p;_p+=n;
	double* dvec_n =_p;_p+=n;
	double* Z =_p;_p+=n;
	double* Wabs =_p;_p+=n;
	double* eta_old =_p;_p+=n;
	double* beta_old =_p;_p+=(p+1);

	
// LOQO control arguments should arrive in alphabetical order - dangerous practice		
	double* primal =_p;_p+=n;
    double* lo = _p;_p+=n;    // 0 <= x
    double* up = Wabs;	      // x <= Wabs
	double* A = _p;_p+=n;
	double dual = lqcontrol->inf;
	double constr_vec1 = 0.0; // not an array since num_constr = 1
	double constr_vec2 = 0.0; // not an array since num_constr = 1
	lqarg->n = n;
	lqarg->m = 1; // number of constraints
	lqarg->H = H;
	lqarg->c = c;
	lqarg->b = &constr_vec1;
	lqarg->r = &constr_vec2;	
	lqarg->A = A;
	lqarg->primal = primal;
	lqarg->lo = lo;
	lqarg->up = up;
	lqarg->double_buffer_size = loqo_buffer_size;
	lqarg->double_buffer = loqo_buffer;
	lqarg->dual = &dual;
	lqarg->primal_obj = lqcontrol->inf;
	lqarg->dual_obj = 0.5*lqcontrol->inf;	
	

	int n0 = 0;
	int n1 = 0;
	for(i = 0;i < n;i++) if(y[i] == 1.0) n1++;else n0++;
	for(i = 0;i < n;i++) Z[i] = SIGN(w[i]);
    for(i = 0;i < n;i++) Wabs[i] = ABS(w[i]);	
	
	
	double* K = Q;
	// int kernel = (int)arg->kernel;getKernel(&n,&p,X,&n,&p,X,&kernel,&arg->kernel,Q);	
	getK(&karg,kernel,K);
	dxd(&n,Z,K,Z,Q); // since Z is n x 1, we can overwrite K=Q with Q=diag(Z)%*%K%*%diag(Z) 	
	dxd(&n,y,Q,y,H);
	double lambda2 = lambda*lambda;for(i = 0;i < (n*n);i++)H[i] /= lambda2;
	for(i = 0;i < n;i++) A[i] = y[i] / lambda;//constr_mat _{m x n}
	for(i = 0;i < n;i++) c[i] = -1.0 / lambda;//linear term _{n x 1}
	// Rprintf("Q=\n");print_matrix(n,n,Q);Rprintf("\n");
	// Rprintf("H=\n");print_matrix(n,n,H);Rprintf("\n");


    if(param->beta_init) memcpy(beta_old,param->beta_init,(p + 1) * sizeof(double));
	// treat intercept separately as x does not contain intercept
	if(!(param->eta_init)) {
		if(kernel->type == LINEAR){
			//eta.init <- Z * (X %*% beta.init[-1])
			matprod(X,&n,&p,beta_old + 1,&p,&ione,eta_old);
			for(i = 0;i < n;i++) eta_old[i] *= Z[i];
			if(param->is_intercept)
				for(i = 0;i < n;i++) 
					eta_old[i] += Z[i] * beta_old[0];
		} else{
			; //eta_old is already set to 0 by calloc;
		}
	}else{
		memcpy(eta_old,param->eta_init,(size_t)n * sizeof(double));
	}
	
    //val.init = sum(Wabs*hinge(1-y*eta.init))+lambda*sum(beta.init[-1]^2)/2
	double* hinge = dvec_n;
	for(i = 0;i < n;i++) hinge[i] = HINGE(1.0 - y[i]*eta_old[i]);
	double val_old = 0.0;for(i = 0;i < n;i++) val_old += Wabs[i] * hinge[i];
	if(lambda != 0.0){
		double pen = 0.0;
		for(int j = 0;j < p;j++)pen += beta_old[j+1] * beta_old[j+1];
		pen *= 0.5 * lambda;
		val_old += pen;		
	}
	double b0_pred = 0.0;	
	////////////////////////////////////////////////////////////////////////////////////////
	
    ////////////////////////////////////////////////////////////////////////////////////////
				
		loqo(lqarg,lqcontrol);		
		
		if((control->verbose > 0) && lqarg->status){
			Rprintf("LOQO did not converge at iteration 1, returned with status %i: %s, error code %i\n",
				lqarg->status,LOQO_STATUS_NAMES[lqarg->status],lqarg->error_code);
		}
		fit->convergence = lqarg->status;
		// allow iteration to continue if convergence is slow (maxit is reached) and let rts convergence decide
		// if((status != LOQO_CONVERGED) && (status != LOQO_SLOW_CONVERGENCE)){	
			// break;
		// }

		// alpha.pred = primal/lambda
		// Q %*% (y*alpha.pred)
		// Q %*% alpha.pred
		for(i = 0;i < n;i++) alpha_pred[i] = primal[i] / lambda;	
		for(i = 0;i < n;i++) y_alpha_pred[i] = y[i] * alpha_pred[i];
		matprod(Q,&n,&n,y_alpha_pred,&n,&ione,Qya_pred);
		matprod(Q,&n,&n,alpha_pred,&n,&ione,Qa_pred);		
		
					
		b0_pred = 0.0;
		int nSV = 0;
		int nBSV = 0;
		int nFSV = 0;
		for(i = 0;i < n;i++)alpha[i] = primal[i] / Wabs[i]; // primal[i] [lo[i],up[i] -> [0,1]
		for(i = 0;i < n;i++){
			int left = (int)(threshold < alpha[i]);	
			int right = (int)(alpha[i] < 1.0 - threshold);	
			nSV  += left;
			nBSV += !right;
			int isFSV = left && right;
			nFSV += isFSV; 
			if(isFSV && (param->is_intercept)) b0_pred += (1.0/y[i] - Qya_pred[i]) / Z[i];		
		}			
		
		if(param->is_intercept){
			if(nFSV >= 2){
				b0_pred /= (double)nFSV;
			}else{				
				// top 5 closest to thresholded-bound
				// top_k_index <- order(apply(cbind(alpha,1.0-alpha),1,min),decreasing=TRUE)[1:min(5,n)] 			
				// top_k_alpha will hold the top 5 a's instead of the indeces (in alpha) of the top 5 a's				
				double top_k_alpha[MIN(5,n)]; 
				int top_k_index[MIN(5,n)];
				for(int j = 0;j < MIN(5,n);j++)top_k_alpha[j] = -DBL_MAX;
				for(i = 0;i < n;i++){
					double a = MIN(alpha[i],1.0-alpha[i]);					
					int ax = i;
					for(int j = 0;j < MIN(5,n);j++){
						if(a >= top_k_alpha[j]){
							double tmp = top_k_alpha[j];
							top_k_alpha[j] = a;
							a = tmp;
							int tx = top_k_index[j];
							top_k_index[j] = ax;
							ax = tx;
						}
					}			
				}
				if(control->verbose > 2){
					Rprintf("threshold free alpha:\n");
					Rprintf("index       :");for(int j = 0;j < MIN(5,n);j++)Rprintf("%i ",top_k_index[j]);Rprintf("\n");
					Rprintf("top alpha   :");for(int j = 0;j < MIN(5,n);j++)Rprintf("%.6f ",top_k_alpha[j]);Rprintf("\n");
					Rprintf("alpha[index]:");for(int j = 0;j < MIN(5,n);j++)Rprintf("%.6f ",alpha[top_k_index[j]]);Rprintf("\n");
				}
				b0_pred = 0.0;
				for(int j = 0;j < MIN(5,n);j++)		
					b0_pred += (1.0/y[top_k_index[j]] - Qya_pred[top_k_index[j]]) / Z[top_k_index[j]];
				b0_pred /= (double)(MIN(5,n));
			}
		}		
		beta_new[0] = b0_pred;
		// Rprintf("b0.pred=%.10f\n",b0_pred);
		// Rprintf("Q       ");for(i = 0;i < n*n;i++)Rprintf("%.6f ",Q[i]);Rprintf("\n");
		// Rprintf("Qya_pred");for(i = 0;i < n;i++)Rprintf("%.6f ",Qya_pred[i]);Rprintf("\n");
		// Rprintf("Qa_pred ");for(i = 0;i < n;i++)Rprintf("%.6f ",Qa_pred[i]);Rprintf("\n");
		// Rprintf("y       ");for(i = 0;i < n;i++)Rprintf("%.6f ",y[i]);Rprintf("\n");
		// Rprintf("z       ");for(i = 0;i < n;i++)Rprintf("%.6f ",Z[i]);Rprintf("\n");
		// Rprintf("a       ");for(i = 0;i < n;i++)Rprintf("%.6f ",alpha[i]);Rprintf("\n");
		
		
		
		// eta.new = Z * b0.pred + Qa.pred
        // val.new = sum(Wabs * hinge(1-y*eta.new))
        // pen = drop(crossprod(alpha.pred*y, Q %*% (alpha.pred*y)))/2*lambda
		
		for(i = 0;i < n;i++) eta_new[i] = b0_pred * Z[i] + Qa_pred[i];
		hinge = dvec_n;
		for(i = 0;i < n;i++) hinge[i] = HINGE(1.0 - y[i]*eta_new[i]);
		double val_new = 0.0;for(i = 0;i < n;i++) val_new += Wabs[i] * hinge[i];
		// Rprintf("eta_new ");for(i = 0;i < n;i++)Rprintf("%.4e ",eta_new[i]);Rprintf("\n");
		

		double pen;crossprod(y_alpha_pred,&n,&ione,Qya_pred,&n,&ione,&pen); 
		pen *= 0.5;
		pen *= lambda;
					    	
       // check stopping criteria and udpate eta.init
		double delta = DBL_MAX;
        // if ((KERNEL_TYPE)(*kernel) == LINEAR){
            // crossprod(X,&n,&p,y_alpha_pred,&n,&ione,beta_new + 1);
            // // delta=max(abs((beta.new[-1]-beta.init[-1])/beta.init[-1])) # skip intercept			
			// delta = ABS((beta_new[0]-beta_old[0]) / beta_old[0]);
			// for(i = 0;i < p;i++) delta = MAX(delta,ABS((beta_new[i+1]-beta_old[i+1]) / beta_old[i+1]));

			// if(verbose > 0){Rprintf(", beta");for(i = 0;i < p+1;i++)Rprintf(" %.4f ",beta_new[i]);Rprintf("delta %.4f\n",delta);}
            // if(delta < tol) {
                // fit->convergence = 0;
                // break;
            // } else {
				// memcpy(beta_old,beta_new,(p + 1) * sizeof(double));
            // }
        // } else 
		{
            val_new += pen;
            delta = val_new - val_old;
            if(control->verbose > 1){ 
				double saturation = 0.0;
				for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
				saturation /= (double)n;
				Rprintf("1 '%s' saturated %.3e loss %.3e pen %.3e delta %.3e\n",
						LOQO_STATUS_NAMES[lqarg->status],saturation,val_new,pen,delta);			
			}
            // if (ABS(delta) < control->tol) {
                // fit->convergence = 0;
                // break;
            // } else {
                // val_old = val_new;
            // }
        }	
		memcpy(eta_old,eta_new,n * sizeof(double));		
	
	if (control->verbose > 0) Rprintf("\n");			

	// Note that there is no division by 'lambda' here as opposed to rtsw()
    // Z.pred = Z
    // Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))
	double* Q_pred = Q;
	// double* Z_pred = Z;
	// getKernel(&n,&p,X,&n,&p,X,kernel,kernel,Q_pred);	
	// dxd(&n,Z_pred,Q_pred,Z_pred,Q_pred); 
		
    // linear.combination = b0.pred + Q.pred %*% (alpha.pred*y)
	double* linear_combination = fit->linear_combination;	
	matprod(Q_pred,&n,&n,y_alpha_pred,&n,&ione,linear_combination);
	for(i = 0;i < n;i++)linear_combination[i] += b0_pred;
	for(i = 0;i < n;i++)
		if(EQ(linear_combination[i],0.0,DBL_EPSILON))
			linear_combination[i] = 0.0;	

    // beta.new = c(b0.pred,drop(crossprod(X, alpha.pred * y)))
	if(kernel->type == LINEAR){
		crossprod(X,&n,&p,y_alpha_pred,&n,&ione,beta_new + 1);
	}
	if(control->verbose == 1){ 
		double saturation = 0.0;
		for(i = 0;i < n;i++)if((eta_new[i] > 0.5/s) || (eta_new[i] < -0.5/s) )saturation++;
		saturation /= (double)n;
		Rprintf("1 '%s' saturated \n",LOQO_STATUS_NAMES[lqarg->status],saturation);			
	}
	// these are aliases only
	memcpy(&fit->param,param,sizeof(RTS_PARAM));				
	memcpy(&fit->kernel,kernel,sizeof(KERNEL_PARAM));
	memcpy(&fit->data,data,sizeof(RTS_MODEL_FRAME));							
	
	if(internal_buffer)free(buffer);
}

void R_rtswt_svm(
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta)
{
	RTS_FIT Fit;memset(&Fit,0,sizeof(RTS_FIT));
	KERNEL_PARAM Kernel;memset(&Kernel,0,sizeof(KERNEL_PARAM));	
	RTS_MODEL_FRAME Data;memset(&Data,0,sizeof(RTS_MODEL_FRAME));
	RTS_MODEL_FRAME* data = &Data;
	RTS_PARAM Param;memset(&Param,0,sizeof(RTS_PARAM));
	RTS_CONTROL* control = &Param.control;	
	LOQO_CONTROL* c = &Param.loqo_control;
	*convergence = 1; //bad
	
// control arguments should arrive in alphabetical order - dangerous practice	
	
	// rts control
	// rts control
	control->maxit = (int)rts_control[0];
	control->tol = rts_control[1];
	control->verbose = (int)rts_control[2];
	
	// quadratic optimizer control
	c->bound = loqo_control[0]; 
	c->inf = loqo_control[1];
	c->margin = loqo_control[2]; 
	c->maxiter = (int)loqo_control[3]; 
	c->sigfig_max = loqo_control[4];
	c->lapack_solver = (LAPACK_SOLVER)loqo_control[5];
	c->verbose = (int)loqo_control[6];	
	
	data->nrx = *nrx;
	data->ncx = *ncx;
	data->X   = X;
	data->y = y;
	data->w = w;
	
	Param.s = *s;
	Param.lambda = *lambda;
	Param.threshold = *threshold;
	
	Param.is_intercept = *is_intercept;
	if(*is_eta){
		Param.neta = *nrx;
		Param.eta_init = eta_init;
	}
	if(*is_beta){
		Param.nbeta = *ncx + 1;
		Param.beta_init = beta_init;
	}

	Kernel.type = (KERNEL_TYPE)*kernel;
	Kernel.param = *kernel_param;
	
	Param.aux_buffer_size = *aux_buffer_size;
	if(Param.aux_buffer_size)Param.aux_buffer = aux_buffer;
	
	Fit.beta = beta;
	Fit.alpha_pred = alpha_pred;
	Fit.linear_combination = linear_combination;
	Fit.eta = eta;
	
	rtswt_svm(&Data,&Param,&Kernel,&Fit);	
	
	*convergence = Fit.convergence;	
}

// # The R function implemented above:
// kernlab.SVW1=function (formula, dat, W, s = 1, lambda = 1, kernel = "linear",
    // para = NULL, start.method = "rlogit", eta0.init = NULL,
    // maxit = 100, tol = 1e-03, verbose = F, minQuad.tol = NULL, cost=0, beta.init,arg=0)
// {

    // #b0.seq=seq(-2,2,0.01)
    // time0 = Sys.time()
    // tmp = model.frame(formula, dat)

    // X = model.matrix(formula, tmp)[, -1, drop = F]  ###
    // y = get_all_vars(formula, dat)[, 1, drop = F]
    // n = nrow(X)  ###
    // TT = dat$TT


    // ### get rid of points where W=0
    // #W<--(y*TT/sum(TT==1)-y*(1-TT)/sum(TT==0)+cost/n)
    // oo<-W!=0
    // W<-W[oo]
    // X<-as.matrix(X[oo,])
    // y<-unlist(y)[oo]




    // Z<-ifelse(W>0,1,-1)
    // Wabs<-abs(W)   # absolute value for W
    // Wm<-diag(Wabs)
    // Zm<-diag(Z)
    // kernel = substr(kernel, 1, 1)
    // if (kernel != "l" & is.null(para))
        // stop("kernel parameter not set")
    // K = getK(X, kernel, para)
    // if (!is.double(K))
        // K <- as.double(K)
    // #Q = Zm %*% K %*% t(Zm)
    // Q = DXD(Zm,K,t(Zm))

    // if (is.null(eta0.init)) {

        // if (kernel == "l"){
          // eta.init = Z * (cbind(1,X) %*% beta.init)
        // } else
        // eta.init=rep(0,sum(oo))
    // } else eta.init=eta0.init

    // val.init = sum(Wabs*hinge(1-y*eta.init))+lambda*sum(beta.init[-1]^2)/2

    // print(val.init)

    // iter = 0
    // converge = 0

     // b = rep(1/lambda,length(Wabs))
     // H<-diag(y)%*%Q%*%diag(y)/lambda^2
     // c <- matrix(-b)
     // A <- t(y/lambda)
     // bb <- 0
     // l <- matrix(rep(0,length(Wabs)))
     // u <- Wabs
     // r <- 0

    // while (TRUE) {
        // iter = iter + 1
        // if (iter > maxit) {
            // converge = 1
            // break
        // }


        // fit <- ipop(c,H,A,bb,l,u,r)

        // if (how(fit) != 'converged')
            // print("minQuad does not converge in Iter " %+% iter)
        // if (verbose)
            // cat(iter)
        // pp<-primal(fit)
        // #alpha.pred = dh2.deta * Wabs + s * pp * Wabs
        // alpha.pred = pp/lambda
       // # eta.new = Q %*% alpha.pred

        // ##### typical support vector machine way to get b0
        // if (arg==0){
            // b0.pred = 1/Z * (1/y - Q %*% (y*alpha.pred))
            // oo.sp = pp/Wabs>1e-4 & pp/Wabs<1-1e-4

            // if (sum(oo.sp)<2) {
                 // pp.min<-apply(cbind(pp/Wabs,1-pp/Wabs),1,min)
                 // oo.sp=order(pp.min,decreasing=TRUE)[1:5]
            // }

            // b0.pred = mean(b0.pred[oo.sp])
        // }

        // #b0.pred=beta.init[1]
        // eta.new = Z * b0.pred + Q %*% alpha.pred
        // val.new = sum (Wabs * hinge(1-y*eta.new))
        // pen = drop(crossprod(alpha.pred*y, Q %*% (alpha.pred*y)))/2*lambda

        // if (verbose)
            // cat(", saturated ", signif(mean(eta.new > 0.5/s |
                // eta.new < -0.5/s), 3))
        // # if (kernel == "l") {
            // # #beta.new = drop(crossprod(X.diff, alpha.pred)/lambda)
            // # beta.new = drop(crossprod(X, alpha.pred * y))
            // # beta.new = c(b0.pred, beta.new)
            // # #if (verbose)
             // # #   cat(", beta[1] ", signif(beta.new[1], 4), ", beta[-1]/beta[1] ",
             // # #     signif(beta.new[-1]/beta.new[1], 4))
            // # if (verbose)
            // # #cat(", beta[1] ", signif(beta.new[1], 4), ", beta[-1]/beta[1] ",
            // # #signif(beta.new[-1]/beta.new[1], 4))
            // # cat(", beta[1] ", signif(beta.new[1], 4), ", beta[-1] ",
            // # signif(beta.new[-1], 4))

            // # if (verbose) {
                // # cat(", loss ", val.new, ", pen ", pen)
            // # val.new = val.new + pen
                // # cat(", val.new ", val.new)
            // # }
          // # #  val.new=sum(Wabs* ramp.f(Z*(cbind(1,X) %*% beta.new),s))
// # #            pen = lambda*beta.new[2]^2/2
// # #            val.new=val.new+pen
// # #            cat(", val.new", signif(val.new,4))
// # #
            // # delta = max(abs((beta.new - beta.init)/beta.init))
            // # if (delta < tol) {
                // # converge = 0
                // # break
            // # }
            // # else {
                // # beta.init = beta.new
                // # eta.init = eta.new
            // # }
        // # }
        // # else 
		// {
            // if (verbose)  cat(", loss ", val.new, ", pen ", pen)
            // val.new = val.new + pen
            // delta = val.new - val.init
            // if (abs(delta) < tol) {
                // converge = 0
                // break
            // }
            // else {
                // val.init = val.new
                // eta.init = eta.new
            // }
        // }
        // if (verbose)
            // cat("\n")
    // }
    // if (verbose)
        // cat("\n")

    // Z.pred = Z
    // #Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))/lambda
    // Q.pred = diag(Z.pred)%*%getK(X, kernel, para)%*%t(diag(Z.pred))

    // #linear.combination = Q.pred %*% alpha.pred

    // linear.combination = b0.pred + Q.pred %*% (alpha.pred*y)
    // result = list(converge = converge, iteration = iter, 
		// b0=b0.pred,beta = drop(beta.new[-1]),coefficients = drop(beta.new), 
		// W=W,X=X,y=y,
        // alpha.pred = drop(alpha.pred), linear.combination = drop(linear.combination),
        // last.minQuad.fit = fit,
      // #  training.auc = fast.auc(drop(linear.combination),
      // #      c(rep(1, n1), rep(0, n2))),
      // #training.theta = caltheta(eta=drop(linear.combination),oo=oo,Y.f=dat$y,TT.f=dat$TT),
      // #      X = X, y = y, lambda = lambda,
        // kernel = kernel, para = para, formula = formula,lambda=lambda)
    // class(result) = c("rauc", class(result))
    // # if (kernel == "l") {
        // # result$coefficients = drop(t(t(alpha.pred) %*% X.diff)/lambda)
        // # result$coefficients = drop(crossprod(X, alpha.pred * y))
        // # result$coefficients = beta.new

    // # }
    // if (result$converge != 0) {
        // result$coefficients = rep(NA, ncol(X))
        // result$alpha.pred = NA
        // result$linear.combination = NA
    // }
    // return(result)
// }
